"""Module to create dummy Scan objects for testing and simulation purposes."""

import numpy as np

from zea.beamform.delays import compute_t0_delays_focused, compute_t0_delays_planewave
from zea.probes import Probe
from zea.scan import Scan


def _get_linear_probe():
    """Returns a probe for ultrasound simulation tests."""
    n_el = 128
    aperture = 30e-3
    probe_geometry = np.stack(
        [
            np.linspace(-aperture / 2, aperture / 2, n_el),
            np.zeros(n_el),
            np.zeros(n_el),
        ],
        axis=1,
    )

    return Probe(
        probe_geometry=probe_geometry,
        center_frequency=2.5e6,
        sampling_frequency=10e6,
    )


def _get_phased_array_probe():
    """Returns a probe for ultrasound simulation tests."""
    n_el = 80
    aperture = 20e-3
    probe_geometry = np.stack(
        [
            np.linspace(-aperture / 2, aperture / 2, n_el),
            np.zeros(n_el),
            np.zeros(n_el),
        ],
        axis=1,
    )

    return Probe(
        probe_geometry=probe_geometry,
        center_frequency=3.12e6,
        sampling_frequency=12.5e6,
    )


def _get_n_ax(ultrasound_probe):
    """Returns the number of ax for ultrasound simulation tests based on the center
    frequency. A probe with a higher center frequency needs more samples to cover
    the image depth.
    """
    is_low_frequency_probe = ultrasound_probe.center_frequency < 4e6

    if is_low_frequency_probe:
        return 510

    return 1024


def _get_probe(kind) -> Probe:
    if kind == "linear":
        return _get_linear_probe()
    elif kind == "phased_array":
        return _get_phased_array_probe()
    else:
        raise ValueError(f"Unknown probe kind: {kind}")


def _get_constant_scan_kwargs():
    return {
        "lens_sound_speed": 1000,
        "lens_thickness": 1e-3,
        "n_ch": 1,
        "selected_transmits": "all",
        "sound_speed": 1540.0,
        "apply_lens_correction": False,
        "attenuation_coef": 0.0,
    }


def _get_lims_and_gridsize(center_frequency, sound_speed):
    """Returns the limits and gridsize for ultrasound simulation tests."""
    xlims, zlims = (-20e-3, 20e-3), (0, 35e-3)
    width, height = xlims[1] - xlims[0], zlims[1] - zlims[0]
    wavelength = sound_speed / center_frequency
    gridsize = (
        int(width / (0.5 * wavelength)) + 1,
        int(height / (0.5 * wavelength)) + 1,
    )
    return {"xlims": xlims, "zlims": zlims, "grid_size_x": gridsize[0], "grid_size_z": gridsize[1]}


def _get_planewave_scan(ultrasound_probe, grid_type, **kwargs):
    """Returns a scan for ultrasound simulation tests."""
    constant_scan_kwargs = _get_constant_scan_kwargs()
    n_el = ultrasound_probe.n_el
    n_tx = 8

    tx_apodizations = np.ones((n_tx, n_el)) * np.hanning(n_el)[None]
    probe_geometry = ultrasound_probe.probe_geometry

    angles = np.linspace(10, -10, n_tx) * np.pi / 180

    sound_speed = constant_scan_kwargs["sound_speed"]
    t0_delays = compute_t0_delays_planewave(
        probe_geometry=probe_geometry, polar_angles=angles, sound_speed=sound_speed
    )

    # Focus distances can be overriden via kwargs
    if "focus_distances" not in kwargs:
        kwargs["focus_distances"] = np.ones(n_tx) * np.inf

    return Scan(
        n_tx=n_tx,
        n_el=n_el,
        center_frequency=ultrasound_probe.center_frequency,
        sampling_frequency=ultrasound_probe.sampling_frequency,
        probe_geometry=probe_geometry,
        t0_delays=t0_delays,
        tx_apodizations=tx_apodizations,
        element_width=np.linalg.norm(probe_geometry[1] - probe_geometry[0]),
        polar_angles=angles,
        initial_times=np.ones(n_tx) * 1e-6,
        n_ax=_get_n_ax(ultrasound_probe),
        grid_type=grid_type,
        **_get_lims_and_gridsize(ultrasound_probe.center_frequency, sound_speed),
        **constant_scan_kwargs,
        **kwargs,
    )


def _get_multistatic_scan(ultrasound_probe, grid_type, **kwargs):
    n_el = ultrasound_probe.n_el
    n_tx = 8

    tx_apodizations = np.zeros((n_tx, n_el))
    for n, idx in enumerate(np.linspace(0, n_el - 1, n_tx, dtype=int)):
        tx_apodizations[n, idx] = 1
    probe_geometry = ultrasound_probe.probe_geometry

    focus_distances = np.zeros(n_tx)
    t0_delays = np.zeros((n_tx, n_el))

    constant_scan_kwargs = _get_constant_scan_kwargs()

    return Scan(
        n_tx=n_tx,
        n_el=n_el,
        center_frequency=ultrasound_probe.center_frequency,
        sampling_frequency=ultrasound_probe.sampling_frequency,
        probe_geometry=probe_geometry,
        t0_delays=t0_delays,
        tx_apodizations=tx_apodizations,
        element_width=np.linalg.norm(probe_geometry[1] - probe_geometry[0]),
        focus_distances=focus_distances,
        polar_angles=np.zeros(n_tx),
        initial_times=np.ones(n_tx) * 1e-6,
        n_ax=_get_n_ax(ultrasound_probe),
        grid_type=grid_type,
        **_get_lims_and_gridsize(
            ultrasound_probe.center_frequency, constant_scan_kwargs["sound_speed"]
        ),
        **constant_scan_kwargs,
        **kwargs,
    )


def _get_diverging_scan(ultrasound_probe, grid_type, **kwargs):
    """Returns a scan for ultrasound simulation tests."""
    constant_scan_kwargs = _get_constant_scan_kwargs()
    n_el = ultrasound_probe.n_el
    n_tx = 8

    tx_apodizations = np.ones((n_tx, n_el)) * np.hanning(n_el)[None]

    angles = np.linspace(10, -10, n_tx) * np.pi / 180

    sound_speed = constant_scan_kwargs["sound_speed"]
    focus_distances = np.ones(n_tx) * -15e-3
    transmit_origins = np.zeros((n_tx, 3))
    t0_delays = compute_t0_delays_focused(
        transmit_origins=transmit_origins,
        focus_distances=focus_distances,
        probe_geometry=ultrasound_probe.probe_geometry,
        polar_angles=angles,
        sound_speed=sound_speed,
    )
    element_width = np.linalg.norm(
        ultrasound_probe.probe_geometry[1] - ultrasound_probe.probe_geometry[0]
    )

    return Scan(
        n_tx=n_tx,
        n_el=n_el,
        center_frequency=ultrasound_probe.center_frequency,
        sampling_frequency=ultrasound_probe.sampling_frequency,
        probe_geometry=ultrasound_probe.probe_geometry,
        t0_delays=t0_delays,
        tx_apodizations=tx_apodizations,
        element_width=element_width,
        focus_distances=focus_distances,
        transmit_origins=transmit_origins,
        polar_angles=angles,
        initial_times=np.ones(n_tx) * 1e-6,
        n_ax=_get_n_ax(ultrasound_probe),
        grid_type=grid_type,
        **_get_lims_and_gridsize(ultrasound_probe.center_frequency, sound_speed),
        **constant_scan_kwargs,
        **kwargs,
    )


def _get_focused_scan(ultrasound_probe, grid_type, **kwargs):
    """Returns a scan for ultrasound simulation tests."""
    constant_scan_kwargs = _get_constant_scan_kwargs()
    n_el = ultrasound_probe.n_el
    n_tx = 8

    tx_apodizations = np.ones((n_tx, n_el)) * np.hanning(n_el)[None]

    angles = np.linspace(30, -30, n_tx) * np.pi / 180

    sound_speed = constant_scan_kwargs["sound_speed"]
    focus_distances = np.ones(n_tx) * 15e-3
    transmit_origins = np.zeros((n_tx, 3))
    t0_delays = compute_t0_delays_focused(
        transmit_origins=transmit_origins,
        focus_distances=focus_distances,
        probe_geometry=ultrasound_probe.probe_geometry,
        polar_angles=angles,
        sound_speed=sound_speed,
    )
    element_width = np.linalg.norm(
        ultrasound_probe.probe_geometry[1] - ultrasound_probe.probe_geometry[0]
    )

    return Scan(
        n_tx=n_tx,
        n_el=n_el,
        center_frequency=ultrasound_probe.center_frequency,
        sampling_frequency=ultrasound_probe.sampling_frequency,
        probe_geometry=ultrasound_probe.probe_geometry,
        t0_delays=t0_delays,
        tx_apodizations=tx_apodizations,
        element_width=element_width,
        focus_distances=focus_distances,
        transmit_origins=transmit_origins,
        polar_angles=angles,
        initial_times=np.ones(n_tx) * 1e-6,
        n_ax=_get_n_ax(ultrasound_probe),
        grid_type=grid_type,
        **_get_lims_and_gridsize(ultrasound_probe.center_frequency, sound_speed),
        **constant_scan_kwargs,
        **kwargs,
    )


def _get_linescan_scan(ultrasound_probe, grid_type, **kwargs):
    """Returns a scan for ultrasound simulation tests."""
    constant_scan_kwargs = _get_constant_scan_kwargs()
    n_el = ultrasound_probe.n_el
    n_tx = 8

    center_elements = np.linspace(0, n_el + 1, n_tx + 2, dtype=int)
    center_elements = center_elements[1:-1]
    tx_apodizations = np.zeros((n_tx, n_el))
    aperture_size_elements = 24

    # Define subapertures
    transmit_origins = []
    for n, idx in enumerate(center_elements):
        el0 = np.clip(idx - aperture_size_elements // 2, 0, n_el)
        el1 = np.clip(idx + aperture_size_elements // 2, 0, n_el)
        tx_apodizations[n, el0:el1] = np.hanning(el1 - el0)[None]
        transmit_origins.append(ultrasound_probe.probe_geometry[idx])
    transmit_origins = np.stack(transmit_origins, axis=0)

    # All angles should be zero because each line fires straight ahead
    angles = np.zeros(n_tx)

    sound_speed = constant_scan_kwargs["sound_speed"]

    focus_distances = np.ones(n_tx) * 15e-3
    t0_delays = compute_t0_delays_focused(
        transmit_origins=transmit_origins,
        focus_distances=focus_distances,
        probe_geometry=ultrasound_probe.probe_geometry,
        polar_angles=angles,
        sound_speed=sound_speed,
    )
    element_width = np.linalg.norm(
        ultrasound_probe.probe_geometry[1] - ultrasound_probe.probe_geometry[0]
    )

    return Scan(
        n_tx=n_tx,
        n_el=n_el,
        center_frequency=ultrasound_probe.center_frequency,
        sampling_frequency=ultrasound_probe.sampling_frequency,
        probe_geometry=ultrasound_probe.probe_geometry,
        t0_delays=t0_delays,
        tx_apodizations=tx_apodizations,
        element_width=element_width,
        focus_distances=focus_distances,
        transmit_origins=transmit_origins,
        polar_angles=angles,
        initial_times=np.ones(n_tx) * 1e-6,
        n_ax=_get_n_ax(ultrasound_probe),
        grid_type=grid_type,
        **_get_lims_and_gridsize(ultrasound_probe.center_frequency, sound_speed),
        **constant_scan_kwargs,
        **kwargs,
    )


def _get_scan(ultrasound_probe, kind, grid_type="cartesian", **kwargs) -> Scan:
    if kind == "planewave":
        return _get_planewave_scan(ultrasound_probe, grid_type, **kwargs)
    elif kind == "multistatic":
        return _get_multistatic_scan(ultrasound_probe, grid_type, **kwargs)
    elif kind == "diverging":
        return _get_diverging_scan(ultrasound_probe, grid_type, **kwargs)
    elif kind == "focused":
        return _get_focused_scan(ultrasound_probe, grid_type, **kwargs)
    elif kind == "linescan":
        return _get_linescan_scan(ultrasound_probe, grid_type, **kwargs)
    else:
        raise ValueError(f"Unknown scan kind: {kind}")


def get_scan(kind="planewave", probe_kind="linear", grid_type="cartesian", **kwargs) -> Scan:
    """Returns a scan for ultrasound simulation tests."""
    ultrasound_probe = _get_probe(probe_kind)
    return _get_scan(ultrasound_probe, kind, grid_type, **kwargs)



import json
from typing import Dict, List, Union

import keras
import numpy as np
import yaml
from keras import ops

from zea import log
from zea.backend import jit
from zea.config import Config
from zea.func.tensor import (
    vmap,
)
from zea.func.ultrasound import channels_to_complex, complex_to_channels
from zea.internal.core import (
    DataTypes,
    ZEADecoderJSON,
    ZEAEncoderJSON,
    dict_to_tensor,
)
from zea.internal.core import Object as ZEAObject
from zea.internal.registry import ops_registry
from zea.internal.utils import deprecated
from zea.ops.base import (
    Operation,
    get_ops,
)
from zea.ops.tensor import Normalize
from zea.ops.ultrasound import (
    ApplyWindow,
    Demodulate,
    EnvelopeDetect,
    LogCompress,
    PfieldWeighting,
    ReshapeGrid,
    TOFCorrection,
)
from zea.probes import Probe
from zea.scan import Scan
from zea.utils import (
    FunctionTimer,
)


@ops_registry("pipeline")
class Pipeline:
    """Pipeline class for processing ultrasound data through a series of operations."""

    def __init__(
        self,
        operations: List[Operation],
        with_batch_dim: bool = True,
        jit_options: Union[str, None] = "ops",
        jit_kwargs: dict | None = None,
        name="pipeline",
        validate=True,
        timed: bool = False,
    ):
        """
        Initialize a pipeline.

        Args:
            operations (list): A list of Operation instances representing the operations
                to be performed.
            with_batch_dim (bool, optional): Whether operations should expect a batch dimension.
                Defaults to True.
            jit_options (str, optional): The JIT options to use. Must be "pipeline", "ops", or None.

                - "pipeline": compiles the entire pipeline as a single function.
                  This may be faster but does not preserve python control flow, such as caching.

                - "ops": compiles each operation separately. This preserves python control flow and
                  caching functionality, but speeds up the operations.

                - None: disables JIT compilation.

                Defaults to "ops".

            jit_kwargs (dict, optional): Additional keyword arguments for the JIT compiler.
            name (str, optional): The name of the pipeline. Defaults to "pipeline".
            validate (bool, optional): Whether to validate the pipeline. Defaults to True.
            timed (bool, optional): Whether to time each operation. Defaults to False.

        """
        self._call_pipeline = self.call
        self.name = name

        self._pipeline_layers = operations

        if jit_options not in ["pipeline", "ops", None]:
            raise ValueError("jit_options must be 'pipeline', 'ops', or None")

        self.with_batch_dim = with_batch_dim
        self._validate_flag = validate

        # Setup timer
        if jit_options == "pipeline" and timed:
            raise ValueError(
                "timed=True cannot be used with jit_options='pipeline' as the entire "
                "pipeline is compiled into a single function. Try setting jit_options to "
                "'ops' or None."
            )
        if timed:
            log.warning(
                "Timer has been initialized for the pipeline. To get an accurate timing estimate, "
                "the `block_until_ready()` is used, which will slow down the execution, so "
                "do not use for regular processing!"
            )
            self._callable_layers = self._get_timed_operations()
        else:
            self._callable_layers = self._pipeline_layers
        self._timed = timed

        if validate:
            self.validate()
        else:
            log.warning("Pipeline validation is disabled, make sure to validate manually.")

        if jit_kwargs is None:
            jit_kwargs = {}

        self._user_jit_kwargs = jit_kwargs.copy()

        if keras.backend.backend() == "jax" and self.static_params != []:
            existing = jit_kwargs.get("static_argnames", [])
            if isinstance(existing, str):
                existing = [existing]
            jit_kwargs = {
                **jit_kwargs,
                "static_argnames": list(set(existing) | set(self.static_params)),
            }

        self.jit_kwargs = jit_kwargs
        self.jit_options = jit_options  # will handle the jit compilation

        self._logged_difference_keys = False

        # Do not log again for nested pipelines
        for nested_pipeline in self._nested_pipelines:
            nested_pipeline._logged_difference_keys = True

    def needs(self, key) -> bool:
        """Check if the pipeline needs a specific key at the input."""
        return key in self.needs_keys

    @property
    def _nested_pipelines(self):
        return [operation for operation in self.operations if isinstance(operation, Pipeline)]

    @property
    def output_keys(self) -> set:
        """All output keys the pipeline guarantees to produce."""
        output_keys = set()
        for operation in self.operations:
            output_keys.update(operation.output_keys)
        return output_keys

    @property
    def valid_keys(self) -> set:
        """Get a set of valid keys for the pipeline.

        This is all keys that can be passed to the pipeline as input.
        """
        valid_keys = set()
        for operation in self.operations:
            valid_keys.update(operation.valid_keys)
        return valid_keys

    @property
    def static_params(self) -> List[str]:
        """Get a list of static parameters for the pipeline."""
        static_params = []
        for operation in self.operations:
            static_params.extend(operation.static_params)
        return list(set(static_params))

    @property
    def needs_keys(self) -> set:
        """Get a set of all input keys needed by the pipeline.

        Will keep track of keys that are already provided by previous operations.
        """
        needs = set()
        has_so_far = set()
        previous_operation = None
        for operation in self.operations:
            if previous_operation is not None:
                has_so_far.update(previous_operation.output_keys)
            needs.update(operation.needs_keys - has_so_far)
            previous_operation = operation
        return needs

    @classmethod
    def from_default(
        cls,
        beamformer="delay_and_sum",
        num_patches=100,
        baseband=False,
        enable_pfield=False,
        timed=False,
        **kwargs,
    ) -> "Pipeline":
        """Create a default pipeline.

        Args:
            beamformer (str): Type of beamformer to use. Currently supporting,
                "delay_and_sum" and "delay_multiply_and_sum". Defaults to "delay_and_sum".
            num_patches (int): Number of patches for the PatchedGrid operation.
                Defaults to 100. If you get an out of memory error, try to increase this number.
            baseband (bool): If True, assume the input data is baseband (I/Q) data,
                which has 2 channels (last dim). Defaults to False, which assumes RF data,
                so input signal has a single channel dim and is still on carrier frequency.
            enable_pfield (bool): If True, apply PfieldWeighting. Defaults to False.
                This will calculate pressure field and only beamform the data to those locations.
            timed (bool, optional): Whether to time each operation. Defaults to False.
            **kwargs: Additional keyword arguments to be passed to the Pipeline constructor.

        """
        operations = []

        # Add the demodulate operation
        if not baseband:
            operations += [
                ApplyWindow(),
                Demodulate(),
            ]

        # Add beamforming ops
        operations.append(
            Beamform(
                beamformer=beamformer,
                num_patches=num_patches,
                enable_pfield=enable_pfield,
            ),
        )

        # Add display ops
        operations += [
            EnvelopeDetect(),
            Normalize(),
            LogCompress(),
        ]
        return cls(operations, timed=timed, **kwargs)

    def copy(self) -> "Pipeline":
        """Create a copy of the pipeline."""
        return Pipeline(
            self._pipeline_layers.copy(),
            with_batch_dim=self.with_batch_dim,
            jit_options=self.jit_options,
            jit_kwargs=self.jit_kwargs,
            name=self.name,
            validate=self._validate_flag,
            timed=self._timed,
        )

    def reinitialize(self):
        """Reinitialize the pipeline in place."""
        self.__init__(
            self._pipeline_layers,
            with_batch_dim=self.with_batch_dim,
            jit_options=self.jit_options,
            jit_kwargs=self.jit_kwargs,
            name=self.name,
            validate=self._validate_flag,
            timed=self._timed,
        )

    def prepend(self, operation: Operation):
        """Prepend an operation to the pipeline."""
        self._pipeline_layers.insert(0, operation)
        self.reinitialize()

    def append(self, operation: Operation):
        """Append an operation to the pipeline."""
        self._pipeline_layers.append(operation)
        self.reinitialize()

    def insert(self, index: int, operation: Operation):
        """Insert an operation at a specific index in the pipeline."""
        if index < 0 or index > len(self._pipeline_layers):
            raise IndexError("Index out of bounds for inserting operation.")
        self._pipeline_layers.insert(index, operation)
        self.reinitialize()

    @property
    def operations(self) -> List[Union[Operation, "Pipeline"]]:
        """Alias for self.layers to match the zea naming convention"""
        return self._pipeline_layers

    def reset_timer(self):
        """Reset the timer for timed operations."""
        if self._timed:
            self._callable_layers = self._get_timed_operations()
        else:
            log.warning(
                "Timer has not been initialized. Set timed=True when initializing the pipeline."
            )

    def _get_timed_operations(self):
        """Get a list of timed operations."""
        self.timer = FunctionTimer()
        return [self.timer(op, name=op.__class__.__name__) for op in self._pipeline_layers]

    def call(self, **inputs):
        """Process input data through the pipeline."""
        for operation in self._callable_layers:
            try:
                outputs = operation(**inputs)
            except KeyError as exc:
                raise KeyError(
                    f"[zea.Pipeline] Operation '{operation.__class__.__name__}' "
                    f"requires input key '{exc.args[0]}', "
                    "but it was not provided in the inputs.\n"
                    "Check whether the objects (such as `zea.Scan`) passed to "
                    "`pipeline.prepare_parameters()` contain all required keys.\n"
                    f"Current list of all passed keys: {list(inputs.keys())}\n"
                    f"Valid keys for this pipeline: {self.valid_keys}"
                ) from exc
            except Exception as exc:
                raise RuntimeError(
                    f"[zea.Pipeline] Error in operation '{operation.__class__.__name__}': {exc}"
                )
            inputs = outputs
        return outputs

    def __call__(self, return_numpy=False, **inputs):
        """Process input data through the pipeline."""

        if any(key in inputs for key in ["probe", "scan", "config"]) or any(
            isinstance(arg, ZEAObject) for arg in inputs.values()
        ):
            raise ValueError(
                "Probe, Scan and Config objects should be first processed with "
                "`Pipeline.prepare_parameters` before calling the pipeline. "
                "e.g. inputs = Pipeline.prepare_parameters(probe, scan, config)"
            )

        if any(isinstance(arg, str) for arg in inputs.values()):
            raise ValueError(
                "Pipeline does not support string inputs. "
                "Please ensure all inputs are convertible to tensors."
            )

        if not self._logged_difference_keys:
            difference_keys = set(inputs.keys()) - self.valid_keys
            if difference_keys:
                log.debug(
                    f"[zea.Pipeline] The following input keys are not used by the pipeline: "
                    f"{difference_keys}. Make sure this is intended. "
                    "This warning will only be shown once."
                )
                self._logged_difference_keys = True

        ## PROCESSING
        outputs = self._call_pipeline(**inputs)

        ## PREPARE OUTPUT
        if return_numpy:
            # Convert tensors to numpy arrays but preserve None values
            outputs = {
                k: ops.convert_to_numpy(v) if ops.is_tensor(v) else v for k, v in outputs.items()
            }

        return outputs

    @property
    def jit_options(self):
        """Get the jit_options property of the pipeline."""
        return self._jit_options

    @jit_options.setter
    def jit_options(self, value: Union[str, None]):
        """Set the jit_options property of the pipeline."""
        self._jit_options = value
        if value == "pipeline":
            assert self.jittable, log.error(
                "jit_options 'pipeline' cannot be used as the entire pipeline is not jittable. "
                "The following operations are not jittable: "
                f"{self.unjitable_ops}. "
                "Try setting jit_options to 'ops' or None."
            )
            self.jit()
            return
        else:
            self.unjit()

        for operation in self.operations:
            if isinstance(operation, Pipeline):
                operation.jit_options = value
            else:
                if operation.jittable and operation._jit_compile:
                    operation.set_jit(value == "ops")

    def jit(self):
        """JIT compile the pipeline."""
        self._call_pipeline = jit(self.call, **self.jit_kwargs)

    def unjit(self):
        """Un-JIT compile the pipeline."""
        self._call_pipeline = self.call

    @property
    def jittable(self):
        """Check if all operations in the pipeline are jittable."""
        return all(operation.jittable for operation in self.operations)

    @property
    def unjitable_ops(self):
        """Get a list of operations that are not jittable."""
        return [operation for operation in self.operations if not operation.jittable]

    @property
    def with_batch_dim(self):
        """Get the with_batch_dim property of the pipeline."""
        return self._with_batch_dim

    @with_batch_dim.setter
    def with_batch_dim(self, value):
        """Set the with_batch_dim property of the pipeline."""
        self._with_batch_dim = value
        for operation in self.operations:
            operation.with_batch_dim = value

    @property
    def input_data_type(self):
        """Get the input_data_type property of the pipeline."""
        return self.operations[0].input_data_type

    @property
    def output_data_type(self):
        """Get the output_data_type property of the pipeline."""
        return self.operations[-1].output_data_type

    def validate(self):
        """Validate the pipeline by checking the compatibility of the operations."""
        operations = self.operations
        for i in range(len(operations) - 1):
            if operations[i].output_data_type is None:
                continue
            if operations[i + 1].input_data_type is None:
                continue
            if operations[i].output_data_type != operations[i + 1].input_data_type:
                raise ValueError(
                    f"Operation {operations[i].__class__.__name__} output data type "
                    f"({operations[i].output_data_type}) is not compatible "
                    f"with the input data type ({operations[i + 1].input_data_type}) "
                    f"of operation {operations[i + 1].__class__.__name__}"
                )

    def set_params(self, **params):
        """Set parameters for the operations in the pipeline by adding them to the cache."""
        for operation in self.operations:
            operation_params = {
                key: value for key, value in params.items() if key in operation.valid_keys
            }
            if operation_params:
                operation.set_input_cache(operation_params)

    def get_params(self, per_operation: bool = False):
        """Get a snapshot of the current parameters of the operations in the pipeline.

        Args:
            per_operation (bool): If True, return a list of dictionaries for each operation.
                                  If False, return a single dictionary with all parameters combined.
        """
        if per_operation:
            return [operation._input_cache.copy() for operation in self.operations]
        else:
            params = {}
            for operation in self.operations:
                params.update(operation._input_cache)
            return params

    def __str__(self):
        """String representation of the pipeline."""
        operations = []
        for operation in self.operations:
            if isinstance(operation, Pipeline):
                operations.append(f"{operation.__class__.__name__}({str(operation)})")
            else:
                operations.append(operation.__class__.__name__)
        string = " -> ".join(operations)
        return string

    def __repr__(self):
        """String representation of the pipeline."""
        operations = []
        for operation in self.operations:
            if isinstance(operation, Pipeline):
                operations.append(repr(operation))
            else:
                operations.append(operation.__class__.__name__)
        return f"<Pipeline {self.name}=({', '.join(operations)})>"

    @classmethod
    def load(cls, file_path: str, **kwargs) -> "Pipeline":
        """Load a pipeline from a JSON or YAML file."""
        if file_path.endswith(".json"):
            with open(file_path, "r", encoding="utf-8") as f:
                json_str = f.read()
            return pipeline_from_json(json_str, **kwargs)
        elif file_path.endswith(".yaml") or file_path.endswith(".yml"):
            return cls.from_path(file_path, **kwargs)
        else:
            raise ValueError("File must have extension .json, .yaml, or .yml")

    def get_dict(self, compact=True) -> dict:
        """Convert the pipeline to a dictionary.

        Args:
            compact (bool): If True (default), only include
                parameters that differ from their defaults.
                If False, include all parameters for full reproducibility.
        """
        config = {"name": ops_registry.get_name(self)}
        config["operations"] = self._pipeline_to_list(self, compact=compact)

        if compact:
            params = {}
            if not self.with_batch_dim:
                params["with_batch_dim"] = self.with_batch_dim
            if self.jit_options != "ops":
                params["jit_options"] = self.jit_options
            if self._user_jit_kwargs:
                params["jit_kwargs"] = self._user_jit_kwargs
            if params:
                config["params"] = params
        else:
            config["params"] = {
                "with_batch_dim": self.with_batch_dim,
                "jit_options": self.jit_options,
                "jit_kwargs": self._user_jit_kwargs,
            }

        return config

    @staticmethod
    def _pipeline_to_list(pipeline: "Pipeline", compact=True):
        """Convert the pipeline to a list of operations."""
        ops_list = []
        for op in pipeline.operations:
            if isinstance(op, Pipeline):
                ops_list.append(op.get_dict(compact=compact))
            else:
                d = op.get_dict(compact=compact)
                if compact:
                    params = d.get("params", {})
                    # Strip with_batch_dim when it is merely inherited from the pipeline
                    if params.get("with_batch_dim") == pipeline.with_batch_dim:
                        params.pop("with_batch_dim", None)
                    # Strip jit_compile=False when it is implied by pipeline-level JIT
                    if not params.get("jit_compile", True) and pipeline.jit_options in (
                        None,
                        "pipeline",
                    ):
                        params.pop("jit_compile", None)
                    if not params:
                        d.pop("params", None)
                    # Name-only dict → bare string shorthand
                    if list(d.keys()) == ["name"]:
                        d = d["name"]
                ops_list.append(d)
        return ops_list

    @classmethod
    def from_config(cls, config: Dict, **kwargs) -> "Pipeline":
        """Create a pipeline from a dictionary or ``zea.Config`` object.

        Args:
            config (dict or Config): Configuration dictionary or ``zea.Config`` object.
                Must have a ``pipeline`` key with a subkey ``operations``.
            **kwargs: Additional keyword arguments to be passed to the pipeline.

        Example:
            .. doctest::

                >>> from zea import Config, Pipeline
                >>> config = Config(
                ...     {
                ...         "pipeline": {
                ...             "operations": [
                ...                 "identity",
                ...             ],
                ...         }
                ...     }
                ... )
                >>> pipeline = Pipeline.from_config(config)
        """
        return pipeline_from_config(Config(config), **kwargs)

    @classmethod
    def from_path(cls, file_path: str, **kwargs) -> "Pipeline":
        """Create a pipeline from a YAML/config file path.

        Args:
            file_path (str): Path to the config file (local or ``hf://`` URI).
                Must have a ``pipeline`` key with a subkey ``operations``.
            **kwargs: Additional keyword arguments to be passed to the pipeline.

        Example:
            .. doctest::

                >>> from zea import Config, Pipeline
                >>> config = Config(
                ...     {
                ...         "pipeline": {
                ...             "operations": [
                ...                 "identity",
                ...             ],
                ...         }
                ...     }
                ... )
                >>> config.to_yaml("pipeline.yaml")
                >>> pipeline = Pipeline.from_path("pipeline.yaml")
        """
        config = Config.from_path(file_path)
        return pipeline_from_config(config, **kwargs)

    @classmethod
    @deprecated(replacement="Pipeline.from_path")
    def from_yaml(cls, file_path: str, **kwargs) -> "Pipeline":
        """Deprecated. Use :meth:`from_path` instead."""
        return pipeline_from_yaml(file_path, **kwargs)

    @classmethod
    def from_json(cls, json_string: str, **kwargs) -> "Pipeline":
        """Create a pipeline from a JSON string.

        Args:
            json_string (str): JSON string representing the pipeline.
                Must have a ``pipeline`` key with a subkey ``operations``.
            **kwargs: Additional keyword arguments to be passed to the pipeline.

        Example:
        ```python
        json_string = '{"pipeline": {"operations": ["identity"]}}'
        pipeline = Pipeline.from_json(json_string)
        ```
        """
        return pipeline_from_json(json_string, **kwargs)

    def to_config(self, compact=True) -> Config:
        """Convert the pipeline to a `zea.Config` object."""
        return pipeline_to_config(self, compact=compact)

    def to_json(self, compact=True) -> str:
        """Convert the pipeline to a JSON string."""
        return pipeline_to_json(self, compact=compact)

    def to_yaml(self, file_path: str, compact=True) -> None:
        """Convert the pipeline to a YAML file."""
        pipeline_to_yaml(self, file_path, compact=compact)

    @property
    def key(self) -> str:
        """Input key of the pipeline."""
        return self.operations[0].key

    @property
    def output_key(self) -> str:
        """Output key of the pipeline."""
        return self.operations[-1].output_key

    def __eq__(self, other):
        """Check if two pipelines are equal."""
        if not isinstance(other, Pipeline):
            return False

        # Compare the operations in both pipelines
        if len(self.operations) != len(other.operations):
            return False

        for op1, op2 in zip(self.operations, other.operations):
            if not op1 == op2:
                return False

        return True

    def prepare_parameters(
        self,
        probe: Probe = None,
        scan: Scan = None,
        config: Config = None,
        **kwargs,
    ):
        """Prepare Probe, Scan and Config objects for the pipeline.

        Serializes `zea.core.Object` instances and converts them to
        dictionary of tensors.

        Args:
            probe: Probe object.
            scan: Scan object.
            config: Config object.
            **kwargs: Additional keyword arguments to be included in the inputs.

        Returns:
            dict: Dictionary of inputs with all values as tensors.
        """
        # Initialize dictionaries for probe, scan, and config
        probe_dict, scan_dict, config_dict = {}, {}, {}

        config_keys, kwargs_keys = set(), set()
        if config is not None:
            config_keys = set(config.keys())
        kwargs_keys = set(kwargs.keys())

        # Process args to extract Probe, Scan, and Config objects
        if probe is not None:
            assert isinstance(probe, Probe), (
                f"Expected an instance of `zea.probes.Probe`, got {type(probe)}"
            )
            probe_dict = probe.to_tensor(keep_as_is=self.static_params)

        if scan is not None:
            assert isinstance(scan, Scan), (
                f"Expected an instance of `zea.scan.Scan`, got {type(scan)}"
            )
            needs_keys = self.needs_keys - config_keys - kwargs_keys
            scan_dict = scan.to_tensor(include=needs_keys, keep_as_is=self.static_params)

        if config is not None:
            assert isinstance(config, Config), (
                f"Expected an instance of `zea.config.Config`, got {type(config)}"
            )
            config_dict.update(config.to_tensor(keep_as_is=self.static_params))

        # Convert all kwargs to tensors
        tensor_kwargs = dict_to_tensor(kwargs, keep_as_is=self.static_params)

        # combine probe, scan, config and kwargs
        # explicitly so we know which keys overwrite which
        # kwargs > config > scan > probe
        inputs = {
            **probe_dict,
            **scan_dict,
            **config_dict,
            **tensor_kwargs,
        }

        return inputs


@ops_registry("map")
class Map(Pipeline):
    """
    A pipeline that maps its operations over specified input arguments.

    This can be used to reduce memory usage by processing data in chunks.

    Notes
    -----
    - When `chunks` and `batch_size` are both None (default), this behaves like a normal Pipeline.
    - Changing anything other than ``self.output_key`` in the dict will not be propagated.
    - Will be jitted as a single operation, not the individual operations.
    - This class handles the batching.

    For more information on how to use ``in_axes``, ``out_axes``, `see the documentation for
    jax.vmap <https://docs.jax.dev/en/latest/_autosummary/jax.vmap.html>`_.

    Example
    -------
        .. doctest::

            >>> from zea.ops import Map, Pipeline, Demodulate, TOFCorrection

            >>> # apply operations in batches of 8
            >>> # in this case, over the first axis of "data"
            >>> # or more specifically, process 8 transmits at a time

            >>> pipeline_mapped = Map(
            ...     [
            ...         Demodulate(),
            ...         TOFCorrection(),
            ...     ],
            ...     argnames="data",
            ...     batch_size=8,
            ... )

            >>> # you can also map a subset of the operations
            >>> # for example, demodulate in 4 chunks
            >>> # or more specifically, split the transmit axis into 4 parts

            >>> pipeline_mapped = Pipeline(
            ...     [
            ...         Map([Demodulate()], argnames="data", chunks=4),
            ...         TOFCorrection(),
            ...     ],
            ... )
    """

    def __init__(
        self,
        operations: List[Operation],
        argnames: List[str] | str,
        in_axes: List[Union[int, None]] | int = 0,
        out_axes: List[Union[int, None]] | int = 0,
        chunks: int | None = None,
        batch_size: int | None = None,
        **kwargs,
    ):
        """
        Args:
            operations (list): List of operations to be performed.
            argnames (str or list): List of argument names (or keys) to map over.
                Can also be a single string if only one argument is mapped over.
            in_axes (int or list): Axes to map over for each argument.
                If a single int is provided, it is used for all arguments.
            out_axes (int or list): Axes to map over for each output.
                If a single int is provided, it is used for all outputs.
            chunks (int, optional): Number of chunks to split the input data into.
                If None, no chunking is performed. Mutually exclusive with ``batch_size``.
            batch_size (int, optional): Size of batches to process at once.
                If None, no batching is performed. Mutually exclusive with ``chunks``.
        """
        super().__init__(operations, **kwargs)

        if batch_size is not None and chunks is not None:
            raise ValueError(
                "batch_size and chunks are mutually exclusive. Please specify only one."
            )

        if batch_size is not None and batch_size <= 0:
            raise ValueError("batch_size must be a positive integer.")

        if chunks is not None and chunks <= 0:
            raise ValueError("chunks must be a positive integer.")

        if isinstance(argnames, str):
            argnames = [argnames]

        self.argnames = argnames
        self.in_axes = in_axes
        self.out_axes = out_axes
        self.chunks = chunks
        self.batch_size = batch_size

        if chunks is None and batch_size is None:
            log.warning(
                "[zea.ops.Map] Both `chunks` and `batch_size` are None. "
                "This will behave like a normal Pipeline. "
                "Consider setting one of them to process data in chunks or batches."
            )

        def call_item(**inputs):
            """Process data in patches."""
            mapped_args = []
            for argname in argnames:
                mapped_args.append(inputs.pop(argname, None))

            def patched_call(*args):
                mapped_kwargs = [(k, v) for k, v in zip(argnames, args)]
                out = super(Map, self).call(**dict(mapped_kwargs), **inputs)

                # TODO: maybe it is possible to output everything?
                # e.g. prepend a empty dimension to all inputs and just map over everything?
                return out[self.output_key]

            out = vmap(
                patched_call,
                in_axes=in_axes,
                out_axes=out_axes,
                chunks=chunks,
                batch_size=batch_size,
                fn_supports_batch=True,
                disable_jit=not bool(self.jit_options),
            )(*mapped_args)

            return out

        self.call_item = call_item

    @property
    def jit_options(self):
        """Get the jit_options property of the pipeline."""
        return self._jit_options

    @jit_options.setter
    def jit_options(self, value):
        """Set the jit_options property of the pipeline."""
        self._jit_options = value
        if value in ["pipeline", "ops"]:
            self.jit()
        else:
            self.unjit()

    def jit(self):
        """JIT compile the pipeline."""
        self._jittable_call = jit(self.jittable_call, **self.jit_kwargs)

    def unjit(self):
        """Un-JIT compile the pipeline."""
        self._jittable_call = self.jittable_call
        self._call_pipeline = self.call

    @property
    def with_batch_dim(self):
        """Get the with_batch_dim property of the pipeline."""
        return self._with_batch_dim

    @with_batch_dim.setter
    def with_batch_dim(self, value):
        """Set the with_batch_dim property of the pipeline.
        The class handles the batching so the operations have to be set to False."""
        self._with_batch_dim = value
        for operation in self.operations:
            operation.with_batch_dim = False

    def jittable_call(self, **inputs):
        """Process input data through the pipeline."""
        if self._with_batch_dim:
            input_data = inputs.pop(self.key)
            output = ops.map(
                lambda x: self.call_item(**{self.key: x, **inputs}),
                input_data,
            )
        else:
            output = self.call_item(**inputs)

        return {self.output_key: output}

    def call(self, **inputs):
        """Process input data through the pipeline."""
        output = self._jittable_call(**inputs)
        inputs.update(output)
        return inputs

    def get_dict(self, compact=True):
        """Get the configuration of the pipeline."""
        config = super().get_dict(compact=compact)
        config["name"] = "map"

        params = config.get("params", {})
        params["argnames"] = self.argnames
        if not compact or self.in_axes != 0:
            params["in_axes"] = self.in_axes
        if not compact or self.out_axes != 0:
            params["out_axes"] = self.out_axes
        if not compact or self.chunks is not None:
            params["chunks"] = self.chunks
        if not compact or self.batch_size is not None:
            params["batch_size"] = self.batch_size
        config["params"] = params
        return config


@ops_registry("patched_grid")
class PatchedGrid(Map):
    """
    A pipeline that maps its operations over `flatgrid` and `flat_pfield` keys.

    This can be used to reduce memory usage by processing data in chunks.

    For more information and flexibility, see :class:`zea.ops.Map`.
    """

    def __init__(self, *args, num_patches=10, **kwargs):
        super().__init__(*args, argnames=["flatgrid", "flat_pfield"], chunks=num_patches, **kwargs)
        self.num_patches = num_patches

    def get_dict(self, compact=True):
        """Get the configuration of the pipeline."""
        config = super().get_dict(compact=compact)
        config["name"] = "patched_grid"

        params = config.get("params", {})
        params.pop("argnames", None)
        params.pop("chunks", None)
        params["num_patches"] = self.num_patches
        config["params"] = params
        return config


@ops_registry("beamform")
class Beamform(Pipeline):
    """Classical beamforming pipeline for ultrasound image formation.

    Expected input data type is `DataTypes.RF_DATA` which has shape `(n_tx, n_ax, n_el, n_ch)`.

    Will run the following operations in sequence:
    - TOFCorrection (output type `DataTypes.ALIGNED_DATA`: `(n_tx, n_ax, n_el, n_ch)`)
    - PfieldWeighting (optional, output type `DataTypes.ALIGNED_DATA`: `(n_tx, n_ax, n_el, n_ch)`)
    - Sum over channels (DAS)
    - Sum over transmits (Compounding) (output type `DataTypes.BEAMFORMED_DATA`: `(grid_size_z, grid_size_x, n_ch)`)
    - ReshapeGrid (flattened grid is also reshaped to `(grid_size_z, grid_size_x)`)
    """  # noqa: E501

    def __init__(self, beamformer="delay_and_sum", num_patches=100, enable_pfield=False, **kwargs):
        """Initialize a Delay-and-Sum beamforming `zea.Pipeline`.

        Args:
            beamformer (str): Type of beamformer to use. Currently supporting,
                "delay_and_sum" and "delay_multiply_and_sum".
            num_patches (int): Number of patches to split the grid into for patch-wise
                beamforming. If 1, no patching is performed.
            enable_pfield (bool): Whether to include pressure field weighting in the beamforming.

        """

        self.beamformer_type = beamformer
        self.num_patches = num_patches
        self.enable_pfield = enable_pfield

        # for backwards compatibility
        name_mapping = {
            "das": "delay_and_sum",
            "dmas": "delay_multiply_and_sum",
        }
        if beamformer in name_mapping:
            log.deprecated(
                f"Beamformer name '{beamformer}' is deprecated. "
                f"Please use '{name_mapping[beamformer]}' instead."
            )
            self.beamformer_type = name_mapping[beamformer]

        if self.beamformer_type not in ["delay_and_sum", "delay_multiply_and_sum"]:
            raise ValueError(
                f"Unsupported beamformer type: {self.beamformer_type}. "
                "Supported types are 'delay_and_sum' and 'delay_multiply_and_sum'."
            )

        # Get beamforming ops
        beamforming = [
            TOFCorrection(),
            # PfieldWeighting(),  # Inserted conditionally
            get_ops(self.beamformer_type)(),
        ]

        if self.enable_pfield:
            beamforming.insert(1, PfieldWeighting())

        # Optionally add patching
        if self.num_patches > 1:
            beamforming = [
                PatchedGrid(
                    operations=beamforming,
                    num_patches=self.num_patches,
                    **kwargs,
                )
            ]

        # Reshape the grid to image shape
        beamforming.append(ReshapeGrid())

        # Set the output data type of the last operation
        # which also defines the pipeline output type
        beamforming[-1].output_data_type = DataTypes.BEAMFORMED_DATA

        super().__init__(operations=beamforming, **kwargs)

    def __repr__(self):
        """String representation of the pipeline."""
        operations = []
        for operation in self.operations:
            if isinstance(operation, Pipeline):
                operations.append(repr(operation))
            else:
                operations.append(operation.__class__.__name__)
        return f"<Beamform {self.name}=({', '.join(operations)})>"

    def get_dict(self, compact=True) -> dict:
        """Convert the pipeline to a dictionary.

        Unlike Pipeline.get_dict(), this does NOT include the internal
        operations list, since Beamform auto-generates its operations
        from ``beamformer``, ``num_patches``, and ``enable_pfield``.
        """
        config = {"name": "beamform"}

        params = {}
        if not compact or self.beamformer_type != "delay_and_sum":
            params["beamformer"] = self.beamformer_type
        if not compact or self.num_patches != 100:
            params["num_patches"] = self.num_patches
        if not compact or self.enable_pfield:
            params["enable_pfield"] = self.enable_pfield

        # Pipeline-level params
        if compact:
            if not self.with_batch_dim:
                params["with_batch_dim"] = self.with_batch_dim
            if self.jit_options != "ops":
                params["jit_options"] = self.jit_options
            if self._user_jit_kwargs:
                params["jit_kwargs"] = self._user_jit_kwargs
        else:
            params["with_batch_dim"] = self.with_batch_dim
            params["jit_options"] = self.jit_options
            params["jit_kwargs"] = self._user_jit_kwargs

        if params:
            config["params"] = params

        return config


@ops_registry("delay_and_sum")
class DelayAndSum(Operation):
    """Sums time-delayed signals along channels and transmits."""

    def __init__(self, **kwargs):
        super().__init__(
            input_data_type=DataTypes.ALIGNED_DATA,
            output_data_type=DataTypes.BEAMFORMED_DATA,
            **kwargs,
        )

    def call(self, **kwargs):
        """Performs DAS beamforming on tof-corrected input.

        Args:
            tof_corrected_data (ops.Tensor): The TOF corrected input of shape
                `(n_tx, prod(grid.shape), n_el, n_ch)` with optional batch dimension.

        Returns:
            dict: Dictionary containing beamformed_data
                of shape `(prod(grid.shape), n_ch)`
                with optional batch dimension.
        """
        data = kwargs[self.key]

        # Sum over the channels (n_el), i.e. DAS
        beamformed_data = ops.sum(data, -2)
        # Sum over transmits (n_tx), i.e. Compounding
        beamformed_data = ops.sum(beamformed_data, -3)

        return {self.output_key: beamformed_data}


@ops_registry("delay_multiply_and_sum")
class DelayMultiplyAndSum(Operation):
    """Performs the operations for the Delay-Multiply-and-Sum beamformer except the delay.
    The delay should be performed by the TOF correction operation.
    """

    def __init__(self, **kwargs):
        super().__init__(
            input_data_type=DataTypes.ALIGNED_DATA,
            output_data_type=DataTypes.BEAMFORMED_DATA,
            **kwargs,
        )

    def process_image(self, data):
        """Performs DMAS beamforming on tof-corrected input.

        Args:
            data (ops.Tensor): The TOF corrected input of shape `(n_tx, n_pix, n_el, n_ch)`

        Returns:
            ops.Tensor: The beamformed data of shape `(n_pix, n_ch)`
        """

        if not data.shape[-1] == 2:
            raise ValueError(
                "MultiplyAndSum operation requires IQ data with 2 channels. "
                f"Got data with shape {data.shape}."
            )

        # Compute the correlation matrix
        data = channels_to_complex(data)

        data = self._multiply(data)
        data = self._select_lower_triangle(data)
        data = ops.sum(data, axis=(0, 2, 3))

        data = complex_to_channels(data)

        return data

    def _select_lower_triangle(self, data):
        """Select only the lower triangle of the correlation matrix."""
        n_el = data.shape[3]
        mask = ops.ones((n_el, n_el), dtype=data.dtype) - ops.eye(n_el, dtype=data.dtype)
        data = data * mask[None, None, :, :] / 2
        return data

    def _multiply(self, data):
        """Apply the DMAS multiplication step."""
        channel_products = data[:, :, :, None] * data[:, :, None, :]

        data = ops.sign(channel_products) * ops.cast(
            ops.sqrt(ops.abs(channel_products)), data.dtype
        )
        return data

    def call(self, **kwargs):
        """Performs DMAS beamforming on tof-corrected input.

        Args:
            tof_corrected_data (ops.Tensor): The TOF corrected input of shape
                `(n_tx, prod(grid.shape), n_el, n_ch)` with optional batch dimension.

        Returns:
            dict: Dictionary containing beamformed_data
                of shape `(grid_size_z*grid_size_x, n_ch)`
                with optional batch dimension.
        """
        data = kwargs[self.key]

        if not self.with_batch_dim:
            beamformed_data = self.process_image(data)
        else:
            # Apply process_image to each item in the batch
            beamformed_data = ops.map(self.process_image, data)

        return {self.output_key: beamformed_data}


def make_operation_chain(
    operation_chain: List[Union[str, Dict, Config, Operation, Pipeline]],
) -> List[Operation]:
    """Make an operation chain from a custom list of operations.

    Args:
        operation_chain (list): List of operations to be performed.
            Each operation can be:
            - A string: operation initialized with default parameters
            - A dictionary: operation initialized with parameters in the dictionary
            - A Config object: converted to a dictionary and initialized
            - An Operation/Pipeline instance: used as-is

    Returns:
        list: List of operations to be performed.

    Example:
        .. doctest::

            >>> from zea.ops import make_operation_chain, LogCompress
            >>> SomeCustomOperation = LogCompress  # just for demonstration
            >>> chain = make_operation_chain(
            ...     [
            ...         "envelope_detect",
            ...         {"name": "normalize", "params": {"output_range": (0, 1)}},
            ...         SomeCustomOperation(),
            ...     ]
            ... )
    """
    chain = []
    for operation in operation_chain:
        # Handle already instantiated Operation or Pipeline objects
        if isinstance(operation, (Operation, Pipeline)):
            chain.append(operation)
            continue

        assert isinstance(operation, (str, dict, Config)), (
            f"Operation {operation} should be a string, dict, Config object, Operation, or Pipeline"
        )

        if isinstance(operation, str):
            operation_instance = get_ops(operation)()

        else:
            if isinstance(operation, Config):
                operation = operation.serialize()

            params = operation.get("params", {})
            op_name = operation.get("name")
            operation_cls = get_ops(op_name)

            # Check for nested operations at the same level as params
            if "operations" in operation:
                nested_operations = make_operation_chain(operation["operations"])
                # Instantiate pipeline-type operations with nested operations
                if issubclass(operation_cls, Beamform):
                    # some pipelines, such as `zea.ops.Beamformer`, are initialized
                    # not with a list of operations but with other parameters that then
                    # internally create a list of operations
                    operation_instance = operation_cls(**params)
                elif issubclass(operation_cls, Pipeline):
                    # in most cases we want to pass an operations list to
                    # initialize a pipeline
                    operation_instance = operation_cls(operations=nested_operations, **params)
                else:
                    operation_instance = operation_cls(operations=nested_operations, **params)
            else:
                operation_instance = operation_cls(**params)

        chain.append(operation_instance)

    return chain


def pipeline_from_config(config: Config, **kwargs) -> Pipeline:
    """
    Create a Pipeline instance from a Config object.

    The config must have a top-level ``pipeline`` key containing an ``operations`` list.
    """
    if "pipeline" not in config:
        top_keys = list(config.keys()) if hasattr(config, "keys") else []
        raise ValueError(
            f"Cannot build Pipeline: missing top-level 'pipeline' key.\n"
            f"Expected a config with the format:\n"
            f"  pipeline:\n"
            f"    operations:\n"
            f"      - <operation_name>\n"
            f"      - ...\n"
            f"Found top-level keys: {top_keys}"
        )

    # Unwrap the pipeline subsection from a full config
    config = Config(config["pipeline"])

    if "operations" not in config:
        top_keys = list(config.keys()) if hasattr(config, "keys") else []
        raise ValueError(
            f"Cannot build Pipeline: missing 'operations' key.\n"
            f"Expected a config with the format:\n"
            f"  pipeline:\n"
            f"    operations:\n"
            f"      - <operation_name>\n"
            f"      - ...\n"
            f"Found top-level keys: {top_keys}"
        )

    if not isinstance(config.operations, (list, np.ndarray)):
        raise ValueError(
            f"Cannot build Pipeline: 'operations' must be a list, "
            f"got {type(config.operations).__name__}."
        )

    operations = make_operation_chain(config.operations)

    # merge pipeline config without operations with kwargs
    pipeline_config = config.copy()
    pipeline_config.pop("operations")

    kwargs = {**pipeline_config, **kwargs}
    return Pipeline(operations=operations, **kwargs)


def pipeline_from_json(json_string: str, **kwargs) -> Pipeline:
    """
    Create a Pipeline instance from a JSON string.
    """
    pipeline_config = Config(json.loads(json_string, cls=ZEADecoderJSON))
    return pipeline_from_config(pipeline_config, **kwargs)


@deprecated(replacement="Pipeline.from_path")
def pipeline_from_yaml(yaml_path: str, **kwargs) -> Pipeline:  # pragma: no cover
    """
    Create a Pipeline instance from a YAML file.

    .. deprecated::
        Use :meth:`Pipeline.from_path` instead.
    """
    with open(yaml_path, "r", encoding="utf-8") as f:
        pipeline_config = yaml.safe_load(f)
    return pipeline_from_config(Config(pipeline_config), **kwargs)


def _pipeline_to_serializable_dict(pipeline: Pipeline, compact=True) -> dict:
    """Convert a Pipeline to a dict suitable for serialization.

    The output format is ``{"pipeline": {"operations": [...], ...pipeline_kwargs}}``
    which can be loaded back via ``pipeline_from_config``.
    """
    pipeline_dict = {
        "operations": Pipeline._pipeline_to_list(pipeline, compact=compact),
    }

    if compact:
        if not pipeline.with_batch_dim:
            pipeline_dict["with_batch_dim"] = pipeline.with_batch_dim
        if pipeline.jit_options != "ops":
            pipeline_dict["jit_options"] = pipeline.jit_options
        if pipeline._user_jit_kwargs:
            pipeline_dict["jit_kwargs"] = pipeline._user_jit_kwargs
        if pipeline.name != "pipeline":
            pipeline_dict["name"] = pipeline.name
    else:
        pipeline_dict.update(
            {
                "with_batch_dim": pipeline.with_batch_dim,
                "jit_options": pipeline.jit_options,
                "jit_kwargs": pipeline._user_jit_kwargs,
                "name": pipeline.name,
            }
        )

    return {"pipeline": pipeline_dict}


def pipeline_to_config(pipeline: Pipeline, compact=True) -> Config:
    """
    Convert a Pipeline instance into a Config object.
    """
    return Config(_pipeline_to_serializable_dict(pipeline, compact=compact))


def pipeline_to_json(pipeline: Pipeline, compact=True) -> str:
    """
    Convert a Pipeline instance into a JSON string.
    """
    return json.dumps(
        _pipeline_to_serializable_dict(pipeline, compact=compact),
        cls=ZEAEncoderJSON,
        indent=4,
    )


def pipeline_to_yaml(pipeline: Pipeline, file_path: str, compact=True) -> None:
    """
    Convert a Pipeline instance into a YAML file.
    """
    with open(file_path, "w", encoding="utf-8") as f:
        yaml.safe_dump(
            _pipeline_to_serializable_dict(pipeline, compact=compact),
            f,
            indent=4,
        )



"""Measurement operators.

Handles task-dependent operations (A) and noises (n) to simulate a measurement y = Ax + n.

"""

import abc

import numpy as np
from keras import ops

from zea.internal.core import Object
from zea.internal.registry import operator_registry


class Operator(abc.ABC, Object):
    """Operator base class.

    Used to define a generatic operator for a specific task / forward model.

    Examples are denoising, inpainting, deblurring, etc.

    One can derive linear and non-linear operators from this class.

    - Linear operators: y = Ax + n
    - Non-linear operators: y = f(x) + n

    """

    @abc.abstractmethod
    def forward(self, data, *args, **kwargs):
        """Implements the forward operator A: x -> y."""
        raise NotImplementedError

    @abc.abstractmethod
    def transpose(self, data, *args, **kwargs):
        """Implements the transpose (or adjoint) of the operator A^T: y -> x."""
        raise NotImplementedError

    @abc.abstractmethod
    def __str__(self):
        """String representation of the operator."""
        raise NotImplementedError


@operator_registry(name="identity")
class IdentityOperator(Operator):
    """Identity operator class."""

    def forward(self, data):
        return data

    def transpose(self, data):
        return data

    def __str__(self):
        return "y = x"


@operator_registry(name="inpainting")
class InpaintingOperator(Operator):
    """Inpainting operator class.

    Inpainting task is a linear operator that masks the data with a mask.

    Formally defined as:
        y = Ax + n, where A = I * M

    Note that this generally only is the case for min_val = 0.0.
    Since we implement the operator using `ops.where`.

    where I is the identity operator, M is the mask, and n is the noise.
    """

    def __init__(self, min_val=0.0, **kwargs):
        """Initialize the inpainting operator.

        Args:
            min_val: Minimum value for the mask.
            **kwargs: Additional arguments.
        """
        super().__init__(**kwargs)
        self.min_val = min_val

    def forward(self, data, mask):
        # return self.mask * data
        return ops.where(mask, data, self.min_val)

    def transpose(self, data, mask):
        # masking operation is diagonal --> A.T = A
        return self.forward(data, mask)

    def __str__(self):
        return "y = Ax + n, where A = I * M"


@operator_registry(name="fourier_blur")
class FourierBlurOperator(Operator):
    """Fourier-domain blurring operator class.

    Applies blurring by masking high frequencies in the Fourier domain.

    Formally defined as:
        y = F^(-1)(M * F(x))

    where F is the FFT, F^(-1) is the inverse FFT and M is the frequency mask.
    """

    def __init__(self, shape, cutoff_freq=0.5, smooth=True, **kwargs):
        """Initialize the Fourier blur operator.

        Args:
            shape: Shape of the input data (H, W), (H, W, C), or (B, H, W, C).
            cutoff_freq: Cutoff frequency as fraction of Nyquist frequency (0.0 to 1.0).
            smooth: If True, use Gaussian rolloff; otherwise use hard cutoff.
            **kwargs: Additional arguments.
        """
        super().__init__(**kwargs)
        self.cutoff_freq = cutoff_freq
        self.shape = shape

        # Precompute frequency mask
        self.freq_mask = self.make_lowpass_mask(shape=shape, cutoff_freq=cutoff_freq, smooth=smooth)

    def make_lowpass_mask(self, shape, cutoff_freq=0.1, smooth=True):
        """
        Create a low-pass Fourier mask of given shape.
        cutoff: relative frequency radius (0 < cutoff < 0.5)
        smooth: if True, use Gaussian rolloff
        """
        # Accept (H, W), (H, W, C) or (B, H, W, C)
        if len(shape) == 2:
            H, W = shape
        elif len(shape) >= 3:
            H, W = shape[-3], shape[-2]
        else:
            raise ValueError(f"Invalid shape {shape}. Expected (H, W), (H, W, C), or (B, H, W, C).")
        fy = np.fft.fftfreq(H)
        fx = np.fft.fftfreq(W)
        FX, FY = np.meshgrid(fx, fy)
        R = np.sqrt(FX**2 + FY**2)

        if smooth:
            sigma = cutoff_freq / np.sqrt(2 * np.log(2))
            mask = np.exp(-(R**2) / (2 * sigma**2))
        else:
            mask = (R < cutoff_freq).astype(np.float32)

        # Shift DC to top-left to match fft2 conventions
        # mask = np.fft.ifftshift(mask)
        return ops.convert_to_tensor(mask)

    def forward(self, data):
        """Apply Fourier-domain blurring.

        Args:
            data: Input tensor of shape (B, H, W, C)

        Returns:
            Blurred data tensor.
        """
        # Convert to float32 for FFT
        data_real = ops.cast(data, "float32")
        data_imag = ops.zeros_like(data_real)

        # fft2 calculates the 2D FFT on the last two dims, so we want
        # H, W to be at the end
        data_real = ops.transpose(data_real, (0, 3, 1, 2))
        data_imag = ops.transpose(data_imag, (0, 3, 1, 2))

        # Apply FFT - expects tuple (real, imag), returns tuple (real, imag)
        fft_real, fft_imag = ops.fft2((data_real, data_imag))

        # Apply frequency mask to both real and imaginary parts
        mask_real = ops.real(self.freq_mask)  # Extract real part of complex mask
        masked_fft_real = fft_real * mask_real
        masked_fft_imag = fft_imag * mask_real

        # Apply inverse FFT
        blurred_real, blurred_imag = ops.ifft2((masked_fft_real, masked_fft_imag))

        # transpose back to original shape
        blurred_real = ops.transpose(blurred_real, (0, 2, 3, 1))

        # Take real part (imaginary should be ~0 for real input)
        blurred_data = ops.cast(blurred_real, data.dtype)

        return blurred_data

    def transpose(self, data):
        """
        transpose = forward because A^* (F^{-1} M F)^* = (F^* M^* (F^{-1})^*) = F^{-1} M F = A
        i.e. this is a self-adjoint operator.
        """
        return self.forward(data)

    def __str__(self):
        return f"y = F^(-1)(M * F(x)) filter at {self.cutoff_freq}"



"""Main entry point for zea

Run as `zea --config path/to/config.yaml` to start the zea interface.
Or do not pass a config file to open a file dialog to choose a config file.

"""

import argparse
import sys
from pathlib import Path

from zea.visualize import set_mpl_style


def get_parser():
    """Command line argument parser"""
    parser = argparse.ArgumentParser(
        description="Load and process ultrasound data based on a configuration file."
    )
    parser.add_argument("-c", "--config", type=str, default=None, help="path to the config file.")
    parser.add_argument(
        "-t",
        "--task",
        default="view",
        choices=["view"],
        type=str,
        help="Which task to run. Currently only 'view' is supported.",
    )
    parser.add_argument(
        "--skip_validate_file",
        default=False,
        action="store_true",
        help="Skip zea file integrity checks. Use with caution.",
    )
    return parser


def main():
    """main entrypoint for zea"""
    args = get_parser().parse_args()

    set_mpl_style()

    wd = Path(__file__).parent.resolve()
    sys.path.append(str(wd))

    from zea.interface import Interface
    from zea.internal.setup_zea import setup

    config = setup(args.config)

    if args.task == "view":
        cli = Interface(
            config,
            validate_file=not args.skip_validate_file,
        )

        cli.run(plot=True)
    else:
        raise ValueError(f"Unknown task {args.task}, see `zea --help` for available tasks.")


if __name__ == "__main__":
    main()



import random as pythonrandom
from keras import Model, Sequential, Variable, layers, ops, random
import numpy as np

from zea.models.hvae.utils import SoftPlus, GaussianAnalyticalKL


class VAE(Model):
    """
    Hierarchical Variational Autoencoder (HVAE) model class.
    Contains an encoder, decoder, and methods for ELBO computation,
    sampling, and model printing.
    """

    def __init__(self, params):
        """
        Args:
            params (Parameters): Parameters object containing model configuration.
        """
        super().__init__()
        self.params = params
        self.encoder = Encoder(params)
        self.decoder = Decoder(params)

        self.beta_max = params.beta
        self.beta_warmup_epochs = params.beta_warmup_epochs

        beta_start = 0 if self.beta_warmup_epochs > 0 else params.beta
        self.beta = Variable(beta_start, trainable=False, dtype="float32")

        self.loss_fn = params.loss_fn
        self.ndims = params.enc_input_size[0] ** 2 * params.data_width
        self.gradient_smoothing = params.gradient_smoothing
        self.min_mol_logscale = -250
        self.softplus = SoftPlus(0.69314718056)

    def build(self):
        """
        Builds the HVAE model by calling.
        """
        self.encoder.build()
        self.decoder.build()
        _ = self.call(
            random.normal(
                [
                    2,  # Batch_size is at least 2 in case of cr-vae
                    self.params.enc_input_size[0],
                    self.params.enc_input_size[0],
                    self.params.data_width,
                ]
            )
        )
        _ = self.call(
            random.normal(
                [
                    2,  # Batch_size is at least 2 in case of cr-vae
                    self.params.enc_input_size[0],
                    self.params.enc_input_size[0],
                    self.params.data_width,
                ]
            ),
        )

    def get_elbo(self, x, px_z, kl, mask=None):
        """
        Calculates the ELBO given the input data, reconstructed output,
        and KL-divergences.

        Args:
            x (tensor): Input data of shape [B, 256, 256, 3], representing targets.
            px_z (tensor): Output logits of shape [B, 256, 256, 100].
            kl (list): List of KL divergences from each latent block.
            mask (tensor, optional): Binary mask of shape [B, 256, 256, 3] to apply to the loss.

        Returns:
            ELBO, reconstruction loss, KL divergence.
        """
        recon_total = self.loss_fn.call(targets=x, logits=px_z, mask=mask)
        kl_total = ops.zeros(ops.shape(x)[0])

        for kls_stage in kl:
            for kl_block in kls_stage:
                if isinstance(kl_block, (list, tuple)) and len(kl_block) == 2:
                    kl_total += ops.sum(kl_block[0], axis=[1, 2, 3])
                    kl_total -= kl_block[1]
                else:
                    kl_total += ops.sum(kl_block, axis=[1, 2, 3])

        if self.params.channels_out > 1:
            recon_total /= self.ndims * ops.log(2)
            kl_total /= self.ndims * ops.log(2)

        recon_total = ops.mean(recon_total)
        kl_total = ops.mean(kl_total)
        elbo = recon_total + self.beta * kl_total

        return elbo, recon_total, kl_total

    def call(self, x):
        """
        Performs a forward pass through the encoder and decoder,
        returning the reconstructed output, latent samples, and KL divergences.

        Args:
            x (tensor): Input tensor of shape [B, 256, 256, 3].

        Returns:
            px_z (tensor): Output logits of shape [B, 256, 256, 100],
            list of latent samples, list of KL divergences of each layer.

        """
        activations = self.encoder(x)
        px_z, z, kl = self.decoder(activations)
        return px_z, z, kl

    def sample_from_mol(self, logits, t=1.0):
        """
        Samples from a mixture of logistics parameterized by the logits.
        In this implementation, usually converts the 100-channel mixture
        into a 3-frame image.

        Args:
            logits (tensor): Logits of shape [B, 256, 256, 100].
            t (float, optional): Temperature for sampling. Defaults to 1.0.

        Returns:
            x (tensor): Sampled image of shape [B, 256, 256, 3] in [-1, 1].
        """

        # Same thing as in the DiscMixLogisticLoss
        B, H, W, _ = ops.shape(logits)

        # Unpack parameters
        logit_probs = logits[:, :, :, : self.params.num_output_mixtures]  # B, H, W, M
        lg = logits[:, :, :, self.params.num_output_mixtures :]  # B, H, W, M*C*3
        lg = layers.Reshape([H, W, self.params.data_width, 3 * self.params.num_output_mixtures])(
            lg
        )  # B, H, W, C, 3*M

        # sample mixture indicator from softmax
        gumbel_noise = -ops.log(
            -ops.log(
                random.uniform(
                    (B, H, W, self.params.num_output_mixtures), minval=1e-5, maxval=1.0 - 1e-5
                )
            )
        )  # B, H, W, M
        amax = ops.argmax(logit_probs / t + gumbel_noise, axis=-1)  # B, H, W
        lambda_ = ops.cast(
            ops.one_hot(amax, self.params.num_output_mixtures, axis=-1), "float32"
        )  # B, H, W, M
        lambda_ = ops.expand_dims(lambda_, axis=3)  # B, H, W, 1, M

        means = ops.sum(
            lg[:, :, :, :, : self.params.num_output_mixtures] * lambda_, axis=-1
        )  # B, H, W, C

        log_scales = self.min_mol_logscale + self.softplus(
            lg[:, :, :, :, self.params.num_output_mixtures : 2 * self.params.num_output_mixtures]
            - self.min_mol_logscale
        )  # B, H, W, C
        log_scales = ops.sum(log_scales * lambda_, axis=-1)  # B, H, W, C

        coeffs = ops.sum(
            ops.tanh(
                lg[
                    :,
                    :,
                    :,
                    :,
                    2 * self.params.num_output_mixtures : 3 * self.params.num_output_mixtures,
                ]
            )
            * lambda_,
            axis=-1,
        )  # B, H, W, C

        # Sample from logistic
        u = random.uniform((B, H, W, self.params.data_width), minval=1e-5, maxval=1.0 - 1e-5)
        x = means + ops.exp(log_scales) * t * (ops.log(u) - ops.log(1.0 - u))  # B, H, W, C

        # Auto-regressive sampling RGB and clip
        if self.params.data_width == 3:
            x0 = ops.clip(x[:, :, :, 0:1], -1, 1)
            x1 = ops.clip(x[:, :, :, 1:2] + coeffs[:, :, :, 0:1] * x0, -1, 1)
            x2 = ops.clip(
                x[:, :, :, 2:3] + coeffs[:, :, :, 1:2] * x0 + coeffs[:, :, :, 2:3] * x1,
                -1,
                1,
            )
            x = ops.concatenate([x0, x1, x2], axis=-1)  # B, H, W, C
        else:
            x = ops.clip(x, -1, 1)

        return x

    def print_model(self):
        """
        Prints the architecture and parameter counts of the HVAE model, split into stages.
        """
        print("------ Encoder ------")
        print(
            (
                f"first 1x1 conv: [None, {
                    (
                        self.params.enc_input_size[0],
                        self.params.enc_input_size[0],
                        self.params.data_width,
                    )
                }]"
                f" -> [None, {
                    (
                        self.params.enc_input_size[0],
                        self.params.enc_input_size[0],
                        self.encoder.first_conv.weights[0].shape[-1],
                    )
                } ]"
            )
        )
        for stage_num, stage in enumerate(self.encoder.stages.layers):
            kernel_size = self.params.kernelsizes[str(stage.input_size)]
            stage_params = 0
            for block in stage.blocks.layers:
                stage_params += block.count_params()
            stage_params += stage.pool.count_params()
            print(
                (
                    f"stage:{stage_num:2} - input_size:{stage.input_size:3}\
                    - in_width:{stage.in_width:3}"
                    f" - middle_width:{stage.middle_width:3} - #blocks:{stage.num_blocks:2}\
                    - ksize:{kernel_size:2} - #params:{stage_params:,d}"
                )
            )
        print("------ Decoder ------")
        for stage_num, stage in enumerate(self.decoder.stages.layers):
            stage_params = 0
            for block in stage.blocks.layers:
                if self.params.use_depthwise_attention:
                    stage_params += block.queries.count_params()
                    if block.combine_queries:
                        stage_params += block.queries_comb_q.count_params()
                        if not block.first_block:
                            stage_params += block.queries_comb_p.count_params()
                stage_params += block.q.count_params()

                stage_params += block.p.count_params()
                stage_params += block.z_out_f.count_params()
                if not block.last_block:
                    stage_params += block.res.count_params()
                    stage_params += block.z_proj.count_params()
            stage_params += stage.pool.count_params()

            print(
                (
                    f"stage:{self.params.num_stages - stage_num - 1:2}\
                    - input_size:{stage.input_size:3} - in_width:{stage.in_width:3}"
                    f" - middle_width:{stage.middle_width:3} - #layers:{stage.num_blocks:2}\
                    - z_ch:{stage.zdim:3} - #params:{stage_params:,d}"
                )
            )
        output_blocks_params = 0
        for block in self.decoder.output_blocks.layers:
            output_blocks_params += block.count_params()
        print(
            (
                f"--- Output ---\n"
                f"blocks   - input_size:{self.decoder.output_shape[1]:3}\
                - in_width:{self.params.z_out_width:3}\
                - middle_width:{self.params.z_out_middle_width:3}\
                - #blocks:{self.decoder.num_output_blocks:2}\
                - ksize:{self.params.kernelsizes[str(self.params.dec_input_size[-1])]:2}\
                - #params:{output_blocks_params:,d}\n"
                f"last 3x3 conv: [None,\
                {stage.input_size, stage.input_size, self.params.z_out_width}]"
                f" -> [None, \
                {stage.input_size, stage.input_size, self.decoder.last_conv.weights[0].shape[-1]}]"
            )
        )
        print("------  Flows  ------")
        if self.params.flow_type == "none":
            print(f"        {None}")
        else:
            for stage_num, stage in enumerate(self.decoder.stages.layers):
                stage_params = 0
                for block in stage.blocks.layers:
                    if block.use_flow:
                        stage_params += block.flows.count_params()
                if self.params.flow_type == "sylvester":
                    print(
                        f"stage:{self.params.num_stages - stage_num - 1:2}\
                        - input_size:{stage.input_size:3}\
                        - eff_zdim:{stage.zdim * stage.input_size * stage.input_size:5}\
                        - #flows:{stage.num_flows:2} - num_ortho_vecs:{stage.num_ortho_vecs:2}\
                        - flow_in_ch:{stage.flow_in_ch:2} - #params:{stage_params:,d}"
                    )
                else:
                    channels = 0
                    flows_per_level = 0
                    split_first = False
                    n_levels = 0
                    use_flow = False
                    if len(stage.blocks.layers) > 0:
                        if stage.blocks.layers[0].use_flow:
                            channels = stage.blocks.layers[0].flows.sylv_channels
                            flows_per_level = stage.blocks.layers[0].flows.flows_per_level
                            split_first = stage.blocks.layers[0].flows.split_first
                            n_levels = stage.blocks.layers[0].flows.n_levels
                            use_flow = True
                    print(
                        f"stage:{self.params.num_stages - stage_num - 1:2} \
                        - input_size:{stage.input_size:3} \
                        - eff_zdim:{stage.zdim * stage.input_size * stage.input_size:5}\
                        - flow:{use_flow:1} - width: {channels:3}\
                        - #flows_per_level:{flows_per_level}\
                        - splitfirst: {split_first:1}\
                        - n_levels: {n_levels:2} - #params:{stage_params:,d}"
                    )
            print(f"flow_type:      {self.params.flow_type}")
            if self.params.flow_type == "conv_sylvester":
                print(f"spectral_norm:  {self.params.spectral_norm}")

        print("---- Attention parameters ----")
        print(f"spatial_attention:   {self.params.use_spatial_attention}")
        if self.params.use_spatial_attention:
            print(f"s_a_width:           {self.params.enc_sa_width}")
        print(f"depthwise_attention: {self.params.use_depthwise_attention}")
        if self.params.use_depthwise_attention:
            print(f"query_width:         {self.params.query_width}")
            print(f"num_queries:         {self.params.num_queries}\n")

        print("---- Model settings ----")
        print(
            (
                f"init_zeros:     {self.params.init_zeros}\n"
                f"block_act:      {self.params.b_act_name}\n"
                f"pool_act:       {self.params.p_act_name}\n"
                f"group_norm:     {self.params.block_bn}\n"
                f"depthwise_conv: {self.params.depthwise}\n"
                f"model_depth:    {self.params.model_depth}\n"
                f"num_enc_blocks: {sum(self.params.enc_num_blocks)}\n"
                f"num_output_mix: {self.params.num_output_mixtures}\n"
                f"grad_smoothing: {self.params.gradient_smoothing}"
            )
        )


class Decoder(layers.Layer):
    """
    Decoder part of the Hierarchical Variational Autoencoder (HVAE).
    Contains the stages that are separated by resolution levels
    (powers of 2).
    """

    def __init__(self, params):
        super().__init__()
        self.output_shape = [
            1,
            params.dec_input_size[-1],
            params.dec_input_size[-1],
            params.dec_in_width[-1],
        ]
        self.z_features_shape = (
            1,
            params.z_out[0],
            params.z_out[1],
            params.z_out[2],
        )
        self.num_stages = params.num_stages
        self.z_to_features = layers.Conv2D(params.z_out_width, kernel_size=1, padding="same")
        self.model_depth = params.model_depth
        self.num_output_blocks = params.output_blocks
        self.init_shape = [
            params.dec_input_size[0],
            params.dec_input_size[0],
            params.dec_in_width[0],
        ]
        self.init_zeros = params.init_zeros
        if not self.init_zeros:
            self.init_bias = self.add_weight(
                shape=(1, self.init_shape[0], self.init_shape[1], self.init_shape[2]),
                initializer="zeros",
                trainable=True,
            )

        self.use_depthwise_attention = params.use_depthwise_attention

        self.stages = Sequential()
        for num in range(self.num_stages):
            self.stages.add(DecoderStage(params, num))

        self.activation = params.block_activation
        self.output_blocks = Sequential()
        for num in range(self.num_output_blocks):
            self.output_blocks.add(
                Block(
                    input_size=self.output_shape[1],
                    in_width=params.z_out_width,
                    middle_width=params.z_out_middle_width,
                    out_width=params.z_out_width,
                    kernelsize=params.kernelsizes[str(params.dec_input_size[-1])],
                    activation=params.block_activation,
                    bn=params.block_bn,
                    residual=True,
                    zero_last=False,
                    model_depth=params.model_depth,
                    depthwise=params.depthwise,
                    use_attention=params.use_spatial_attention,
                    attention_width=params.dec_sa_width[-1],
                )
            )
        self.last_conv = layers.Conv2D(
            params.channels_out,
            kernel_size=params.kernelsizes[str(params.dec_input_size[-1])],
            padding="same",
        )

    def build(self):
        for dec_stage in self.stages.layers:
            dec_stage.build()
        for out_block in self.output_blocks.layers:
            out_block.build()

    def call(self, activations):
        """
        Creates an output image from the encoder activations.
        args:
            activations: List of encoder activations at different resolutions.
        returns:
            Reconstructed image, latent samples, KL divergences.

        - Starts from an empty tensor or learned bias.
        - Passes through decoder stages with encoder activations.
        - Sums all latents.
        - Passes this latent through output blocks and reconstructs image.

        """
        if self.use_depthwise_attention:
            x, z_stages, kl_stages = self.depthwise_attention_call(activations)
        else:
            z_stages = []
            kl_stages = []

            if self.init_zeros:
                x = ops.zeros_like(activations[-1])
            else:
                b = ops.shape(activations[-1])[0]
                x = ops.repeat(self.init_bias, b, axis=0)

            for dec_stage, act in zip(self.stages.layers, reversed(activations)):
                x, z, kl = dec_stage(x, act)
                z_stages.append(z)
                kl_stages.append(kl)

        z_out = sum(z_stages) / ops.sqrt(self.model_depth)

        px_z = self.activation(self.z_to_features(z_out))
        for out_block in self.output_blocks.layers:
            px_z = out_block(px_z)
        px_z = self.last_conv(px_z)

        return px_z, z_stages, kl_stages

    def call_uncond(self, num_images=16, t=1):
        """
        Generates an image from the prior of every stage.
        """
        if self.init_zeros:
            x = ops.zeros((num_images, self.init_shape[0], self.init_shape[1], self.init_shape[2]))
        else:
            x = ops.repeat(self.init_bias, num_images, axis=0)

        if self.use_depthwise_attention:
            # vp and kp get overwritten in first call, these are placeholders
            vp = ops.zeros(1)
            kp = ops.zeros(1)
            # pack top-down stream as x, vp, kp
            x = (x, vp, kp)

        z_stages = []
        for dec_stage in self.stages.layers:
            x, z = dec_stage.call_uncond(x, t)
            z_stages.append(z)

        z_out = sum(z_stages) / ops.sqrt(self.model_depth)

        px_z = self.activation(self.z_to_features(z_out))
        for out_block in self.output_blocks.layers:
            px_z = out_block(px_z)
        px_z = self.last_conv(px_z)

        return px_z


class Encoder(layers.Layer):
    """
    Encoder part of the Hierarchical Variational Autoencoder (HVAE).
    Contains the stages that are separated by resolution levels.
    """

    def __init__(self, params):
        super().__init__()
        self.input_shape = [
            1,
            params.enc_input_size[0],
            params.enc_input_size[0],
            params.data_width,
        ]
        self.num_stages = params.num_stages
        self.first_conv = layers.Conv2D(
            params.enc_in_width[0],
            kernel_size=params.kernelsizes[str(params.enc_input_size[0])],
            padding="same",
        )

        self.stages = Sequential()
        for num in range(self.num_stages):
            self.stages.add(EncoderStage(params, num))

    def build(self):
        for enc_stage in self.stages.layers:
            enc_stage.build()
        _ = self.call(random.normal(self.input_shape))

    def call(self, x):
        """
        Args:
            x: Input tensor to the encoder.
        Returns:
            List of encoder activations at different resolutions.
        """
        x = self.first_conv(x)
        activations = []
        for enc_stage in self.stages.layers:
            # If there is no depth_wise attention, activations are passed to decoder
            # If there is depth_wise attention, act[0]=v^q, act[1]=k^q
            # every v^q and k^q is at a different resolution,
            # pooling/upsampling happens at attention block

            x, act = enc_stage(x)
            activations += [act]
        return activations


class DecoderStage(layers.Layer):
    def __init__(self, params, stage_num):
        super().__init__()
        self.zdim = params.zdim[stage_num]
        self.input_size = params.dec_input_size[stage_num]
        self.in_width = params.dec_in_width[stage_num]
        self.middle_width = params.dec_middle_width[stage_num]
        self.pool_width = params.dec_pool_width[stage_num]
        self.num_blocks = params.dec_num_blocks[stage_num]

        # attention
        self.use_spatial_attention = params.use_spatial_attention
        self.sa_width = params.dec_sa_width[stage_num]
        self.use_depthwise_attention = params.use_depthwise_attention

        # sylvester flow params
        self.num_flows = params.num_flows[stage_num]
        self.flow_in_ch = params.flow_in_ch[stage_num] if self.num_flows > 0 else 0
        self.num_ortho_vecs = params.num_ortho_vecs[stage_num] if self.num_flows > 0 else 0

        # conv_sylvester flow param
        self.convsylv_channels = params.convsylv_channels[stage_num]
        self.convsylv_flows_per_stage = params.convsylv_flows_per_stage[stage_num]
        self.convsylv_splitfirst = params.convsylv_splitfirst[stage_num]
        self.convsylv_stage_limit = params.convsylv_stage_limit[stage_num]

        self.z_out = params.z_out

        # For the decoder, a block is a DecBlock
        self.blocks = Sequential()
        for i in range(self.num_blocks):
            first_block = False
            last_block = False
            if i == 0 and self.input_size == min(params.dec_input_size):
                first_block = True
            if i == self.num_blocks - 1 and self.input_size == max(params.dec_input_size):
                last_block = True

            self.blocks.add(
                DecBlock(
                    input_size=self.input_size,
                    in_width=self.in_width,
                    middle_width=self.middle_width,
                    zdim=self.zdim,
                    z_out=params.z_out,
                    kernelsize=params.kernelsizes[str(self.input_size)],
                    block_activation=params.block_activation,
                    block_bn=params.block_bn,
                    model_depth=params.model_depth,
                    depthwise=params.depthwise,
                    num_flows=self.num_flows,
                    flow_in_ch=self.flow_in_ch,
                    num_ortho_vecs=self.num_ortho_vecs,
                    flow_type=params.flow_type,
                    spectral_norm=params.spectral_norm,
                    convsylv_channels=self.convsylv_channels,
                    convsylv_flows_per_stage=self.convsylv_flows_per_stage,
                    convsylv_splitfirst=self.convsylv_splitfirst,
                    convsylv_stage_limit=self.convsylv_stage_limit,
                    gradient_smoothing=params.gradient_smoothing,
                    use_spatial_attention=self.use_spatial_attention,
                    sa_width=self.sa_width,
                    use_depthwise_attention=self.use_depthwise_attention,
                    query_width=params.query_width,
                    num_queries=params.num_queries,
                    first_block=first_block,
                    last_block=last_block,
                )
            )
        self.pool = PoolLayer(
            input_size=self.input_size,
            in_width=self.in_width,
            out_width=self.pool_width,
            pool_activation=params.pool_activation,
            unpool=True,
            data_size=params.enc_input_size[0],
        )

    def build(self):
        for dec_block in self.blocks.layers:
            dec_block.build()
        self.pool.build()

    def call(self, x, act):
        B = ops.shape(x[0])[0] if self.use_depthwise_attention else ops.shape(x)[0]
        z_blocks = ops.tile(ops.zeros([1, *self.z_out]), (B, 1, 1, 1))
        kl_blocks = []
        for dec_block in self.blocks.layers:
            x, z, kl = dec_block.call(x, act)
            z_blocks += z
            kl_blocks.append(kl)

        if not self.use_depthwise_attention:
            x = self.pool(x)
        else:
            dec, vp, kp = x
            x = (self.pool(dec), self.attn_pool(vp), self.attn_pool(kp))
        return x, z_blocks, kl_blocks

    def call_uncond(self, x, t=1):
        B = ops.shape(x[0])[0] if self.use_depthwise_attention else ops.shape(x)[0]
        z_blocks = ops.tile(ops.zeros([1, *self.z_out]), (B, 1, 1, 1))
        for dec_block in self.blocks.layers:
            x, z = dec_block.call_uncond(x, t)
            z_blocks += z

        if not self.use_depthwise_attention:
            x = self.pool(x)
        else:
            dec, vp, kp = x
            x = (self.pool(dec), self.attn_pool(vp), self.attn_pool(kp))
        return x, z_blocks


class EncoderStage(layers.Layer):
    def __init__(self, params, stage_num):
        super().__init__()
        self.stage_num = stage_num

        self.input_size = params.enc_input_size[stage_num]
        self.in_width = params.enc_in_width[stage_num]
        self.middle_width = params.enc_middle_width[stage_num]
        self.pool_width = params.enc_pool_width[stage_num]
        self.num_blocks = params.enc_num_blocks[stage_num]
        self.use_spatial_attention = params.use_spatial_attention
        self.sa_width = params.enc_sa_width[stage_num]

        self.use_depthwise_attention = params.use_depthwise_attention
        if self.use_depthwise_attention:
            self.key_width = params.query_width
            self.lw_v = layers.LayerNormalization(axis=[1, 2, 3])

        # For the encoder, a block is just a ResNet Block
        out_width = self.in_width
        self.blocks = Sequential()
        for i in range(self.num_blocks):
            out_width = self.in_width
            # If this is the final block of the stage, we also output the keys k^q_l
            if i == self.num_blocks - 1 and self.use_depthwise_attention:
                out_width += self.key_width

            in_width = self.in_width
            # If this is the first block of the second, third, etc stage, input keys
            # if stage_num > 0 and i == 0 and self.use_depthwise_attention:
            # in_width += self.key_width

            self.blocks.add(
                Block(
                    input_size=self.input_size,
                    in_width=in_width,
                    middle_width=self.middle_width,
                    out_width=out_width,
                    kernelsize=params.kernelsizes[str(self.input_size)],
                    activation=params.block_activation,
                    bn=params.block_bn,
                    residual=True,
                    zero_last=False,
                    model_depth=params.model_depth,
                    depthwise=False,
                    use_attention=self.use_spatial_attention,
                    attention_width=self.sa_width,
                    dwa_enc=self.use_depthwise_attention,
                )
            )

        self.pool = PoolLayer(
            input_size=self.input_size,
            in_width=out_width,
            out_width=self.pool_width,
            pool_activation=params.pool_activation,
            unpool=False,
        )

    def build(self):
        for enc_block in self.blocks.layers:
            enc_block.build()
        self.pool.build()

        in_width = self.in_width
        # if self.stage_num > 0 and self.use_depthwise_attention:
        #     in_width += self.key_width

        _, _ = self.call(random.normal([1, self.input_size, self.input_size, in_width]))

    def call(self, x):
        for enc_block in self.blocks.layers:
            # Only the final block will have x with more channels
            x = enc_block(x)

        # return pooled x for the next stage and the activations for the decoder
        if self.use_depthwise_attention:
            if self.num_blocks > 0:
                vq = x[:, :, :, : self.in_width]  # (B, H, W, in_width)
                vq += ops.gelu(self.lw_v(vq))
                kq = x[:, :, :, self.in_width :]  # (B, H, W, key_width)
            else:
                vq = None
                kq = None
            return self.pool(x), [vq, kq]
        else:
            return self.pool(x), x


class DecBlock(layers.Layer):
    """
    Flow layers and attention are not available in this code snippet.
    This means that all flow-related and attention-related arguments are omitted.
    The functions are still there to allow for compatibility with the code origin.
    """

    def __init__(
        self,
        input_size,
        in_width,
        middle_width,
        zdim,
        z_out,
        kernelsize,
        block_activation,
        block_bn,
        model_depth,
        depthwise,
        num_flows,
        flow_in_ch,
        num_ortho_vecs,
        flow_type,
        spectral_norm,
        convsylv_channels,
        convsylv_flows_per_stage,
        convsylv_splitfirst,
        convsylv_stage_limit,
        gradient_smoothing,
        use_spatial_attention,
        sa_width,
        use_depthwise_attention,
        query_width,
        num_queries,
        first_block,
        last_block,
    ):
        super().__init__()
        self.first_block = first_block
        self.last_block = last_block
        # The blocks cannot have an identical seedgenerator, they also cannot share a single one,
        # this is necessary because of stateless mirrorstrategy
        seed = pythonrandom.randint(0, 1_000_000)
        self.seed_gen = random.SeedGenerator(1337 + seed)

        # Parameters used for building
        self.input_size = input_size
        self.in_width = in_width
        self.middle_width = middle_width
        self.zdim = zdim
        self.model_depth = model_depth

        self.use_spatial_attention = use_spatial_attention
        self.sa_width = sa_width
        self.use_depthwise_attention = use_depthwise_attention

        self.num_flows = num_flows
        self.flow_in_ch = flow_in_ch
        self.num_ortho_vecs = num_ortho_vecs
        self.flow_type = flow_type

        if self.flow_type != "none" and self.num_flows > 0:
            self.use_flow = True
        else:
            self.use_flow = False

        p_out_width = 2 * self.zdim + in_width
        if self.use_depthwise_attention:
            self.query_width = query_width
            self.num_queries = num_queries
            p_out_width += self.query_width

            self.combine_queries = self.num_queries > 1
            if self.combine_queries:
                self.queries_comb_q = layers.Conv2D(self.in_width, kernel_size=1)
                if not self.first_block:
                    self.queries_comb_p = layers.Conv2D(self.in_width, kernel_size=1)

            queries_out = self.query_width * num_queries
            self.gamma_q = Variable(0, trainable=True, dtype="float32")

            if not self.first_block:
                # Also output prior queries
                queries_out *= 2
                self.gamma_p = Variable(0, trainable=True, dtype="float32")

            self.queries = Block(
                input_size=self.input_size,
                in_width=self.in_width,
                middle_width=self.middle_width,
                out_width=queries_out,
                kernelsize=kernelsize,
                activation=block_activation,
                bn=block_bn,
                residual=False,
                zero_last=False,
                model_depth=model_depth,
                depthwise=depthwise,
                use_attention=use_spatial_attention,
                attention_width=sa_width,
            )

        q_out_width = self.zdim * 2
        # Block that takes activations from encoder
        self.q = Block(
            input_size=self.input_size,
            in_width=2 * self.in_width,
            middle_width=self.middle_width,
            out_width=q_out_width,
            kernelsize=kernelsize,
            activation=block_activation,
            bn=block_bn,
            residual=False,
            zero_last=False,
            model_depth=model_depth,
            depthwise=depthwise,
            use_attention=use_spatial_attention,
            attention_width=sa_width,
        )
        self.kl = GaussianAnalyticalKL()

        self.p = Block(
            input_size=self.input_size,
            in_width=self.in_width,
            middle_width=self.middle_width,
            out_width=p_out_width,
            kernelsize=kernelsize,
            activation=block_activation,
            bn=block_bn,
            residual=False,
            zero_last=True,
            model_depth=model_depth,
            depthwise=depthwise,
            use_attention=use_spatial_attention,
            attention_width=sa_width,
        )

        # Residual block for after adding Z
        if not self.last_block:
            self.res = Block(
                input_size=self.input_size,
                in_width=self.in_width,
                middle_width=self.middle_width,
                out_width=self.in_width,
                kernelsize=kernelsize,
                activation=block_activation,
                bn=block_bn,
                residual=True,
                zero_last=False,
                model_depth=model_depth,
                depthwise=depthwise,
                use_attention=use_spatial_attention,
                attention_width=sa_width,
            )
            self.z_proj = layers.Conv2D(in_width, kernel_size=1)

        self.sp = SoftPlus(gradient_smoothing)
        self.z_out_f = layers.Conv2D(z_out[-1], kernel_size=1)
        if z_out[0] > input_size:
            self.z_out_up = layers.UpSampling2D(
                size=z_out[0] // input_size, interpolation="nearest"
            )
        else:
            self.z_out_up = layers.Identity()

    def build(self):
        self.q.build()
        self.p.build()
        if self.use_flow:
            self.flows.build()

        if not self.last_block:
            self.res.build()
            _ = self.z_proj(random.normal([1, self.input_size, self.input_size, self.zdim]))
            weights = self.z_proj.get_weights()
            weights[0] *= np.sqrt(1 / self.model_depth)
            self.z_proj.set_weights(weights)

    def sample(self, x, act):
        # Calculate all attention outputs
        q_out = self.q(ops.concatenate([x, act], axis=-1))
        qm, q_std = ops.split(q_out, 2, axis=3)
        h = None

        # calculate prior (mu_p, variance_p, and residual)
        prior = self.p(x)
        pm, p_std, vpl = (
            prior[:, :, :, : self.zdim],
            prior[:, :, :, self.zdim : self.zdim * 2],
            prior[:, :, :, self.zdim * 2 :],
        )
        x = ops.add(x, vpl)

        q_std = self.sp(q_std)
        p_std = self.sp(p_std)

        noise = random.normal(ops.shape(q_std), seed=self.seed_gen)

        z0 = ops.add(qm, ops.multiply(q_std, noise))

        if self.use_flow:
            zk, log_det_j = self.flows.call(z0, h)
            kl = self.kl.call(qm, q_std, pm, p_std, z0, zk)
            return x, zk, (kl, log_det_j)
        else:
            kl = self.kl.call(qm, q_std, pm, p_std)
            return x, z0, kl

    def sample_uncond(self, x, t=1):
        if self.use_depthwise_attention:
            pass
        else:
            prior = self.p(x)
            pm, p_std, vpl = (
                prior[:, :, :, : self.zdim],
                prior[:, :, :, self.zdim : self.zdim * 2],
                prior[:, :, :, self.zdim * 2 :],
            )
            x = ops.add(x, vpl)

        p_std = self.sp(p_std)
        noise = random.normal(ops.shape(p_std), seed=self.seed_gen)
        z = pm + p_std * noise * t
        return x, z

    def call(self, x, act):
        x, z, kl = self.sample(x, act)
        if self.use_depthwise_attention:
            x, vp, kp = x
        if not self.last_block:
            x = ops.add(x, self.z_proj(z))
            x = self.res(x)
        if self.use_depthwise_attention:
            x = (x, vp, kp)

        z = self.z_out_f(z)
        z = self.z_out_up(z)
        return x, z, kl

    def call_uncond(self, x, t=1):
        x, z = self.sample_uncond(x, t)
        if self.use_depthwise_attention:
            x, vp, kp = x
        if not self.last_block:
            x = ops.add(x, self.z_proj(z))
            x = self.res(x)
        if self.use_depthwise_attention:
            x = (x, vp, kp)

        z = self.z_out_f(z)
        z = self.z_out_up(z)
        return x, z


class Block(layers.Layer):
    """
    This class represents a single ResNet block and all the options it can have.
    Again, self-attention is not implemented in this snippet.
    """

    def __init__(
        self,
        input_size,
        in_width,
        middle_width,
        out_width,
        kernelsize,
        activation,
        bn,
        residual,
        zero_last,
        model_depth,
        depthwise,
        use_attention,
        attention_width,
        dwa_enc=False,
    ):
        super().__init__()
        # Parameters used for building
        self.input_size = input_size
        self.in_width = in_width
        self.middle_width = middle_width
        self.out_width = out_width
        self.zero_last = zero_last
        self.model_depth = model_depth
        self.depthwise = depthwise

        # Parameters used for calling
        self.activation = activation
        self.residual = residual
        self.use_attention = use_attention and (attention_width > 0)
        if self.use_attention:
            self.attention_width = attention_width
        self.dwa_enc = dwa_enc

        # Groupnorm layers
        self.gn1 = layers.GroupNormalization(groups=in_width // 8) if bn else layers.Identity()
        self.gn2 = layers.GroupNormalization(groups=middle_width // 8) if bn else layers.Identity()
        self.gn3 = layers.GroupNormalization(groups=middle_width // 8) if bn else layers.Identity()
        self.gn4 = layers.GroupNormalization(groups=middle_width // 8) if bn else layers.Identity()

        # Convolutional layers
        self.c1 = layers.Conv2D(middle_width, kernel_size=1)
        if depthwise:
            # Depthwise 5x5 convs as used in NVAE
            self.c2 = (
                layers.DepthwiseConv2D(kernel_size=5, padding="same")
                if (self.input_size > 4)
                else layers.Conv2D(middle_width, kernel_size=1)
            )
            self.c3 = (
                layers.DepthwiseConv2D(kernel_size=5, padding="same")
                if (self.input_size > 4)
                else layers.Conv2D(middle_width, kernel_size=1)
            )
        else:
            # 3x3 convs as used in (Efficient)-VDVAE
            self.c2 = layers.Conv2D(middle_width, kernel_size=kernelsize, padding="same")
            self.c3 = layers.Conv2D(middle_width, kernel_size=kernelsize, padding="same")
        kern_init = "zeros" if zero_last else "glorot_uniform"
        self.c4 = layers.Conv2D(out_width, kernel_size=1, kernel_initializer=kern_init)

        # SPATIAL-SELF-ATTENTION for deep attentive variational inference
        if self.use_attention:
            pass

    def build(self):
        _ = self.call(random.normal([1, self.input_size, self.input_size, self.in_width]))

        # Initialize last layer with 1/sqrt(n)
        weights = self.c4.get_weights()
        weights[0] *= np.sqrt(1 / self.model_depth)
        self.c4.set_weights(weights)

    def call(self, x):
        xhat = self.c1(self.activation(self.gn1(x)))
        if self.use_attention:
            xhat = self.attention1(xhat)

        xhat = self.c2(self.activation(self.gn2(xhat)))
        if self.use_attention:
            xhat = self.attention2(xhat)

        xhat = self.c3(self.activation(self.gn3(xhat)))
        xhat = self.c4(self.activation(self.gn4(xhat)))

        if self.dwa_enc and self.in_width < self.out_width:
            out = ops.concatenate(
                [
                    ops.add(xhat[:, :, :, : self.in_width], x),  # vq
                    xhat[:, :, :, self.in_width :],  # kq
                ],
                axis=-1,
            )
        elif self.dwa_enc and self.in_width > self.out_width:
            out = ops.add(xhat, x[:, :, :, : self.out_width])  # vq

        else:
            out = ops.add(x, xhat) if self.residual else xhat
        return out


class PoolLayer(layers.Layer):
    """
    Pooling layer that can perform either:
    AveragePooling2D for the encoder,
    Upsampling2D for the decoder.
    """

    def __init__(self, input_size, in_width, out_width, pool_activation, unpool, data_size=None):
        super().__init__()
        # Building
        self.input_size = input_size
        self.in_width = in_width
        self.out_width = out_width
        self.activation = pool_activation

        # Layers
        if in_width == out_width:
            self.c1 = layers.Identity()
        else:
            self.c1 = layers.Conv2D(out_width, kernel_size=1)

        if unpool:
            self.pool = layers.UpSampling2D(2) if (input_size < data_size) else layers.Identity()
        else:
            self.pool = layers.AveragePooling2D(2) if (input_size > 1) else layers.Identity()

    def build(self):
        _ = self.call(random.normal([1, self.input_size, self.input_size, self.in_width]))

    def call(self, x):
        return self.pool(self.activation(self.c1(x)))



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation

from zea import Scan
from zea.io_lib import save_to_gif


def animate_images(
    images, path, scan: Scan = None, interval=100, cmap="gray", figsize=(5, 4), dpi=80
):
    """Helper function to animate a list of images."""
    if interval <= 0:
        raise ValueError("interval must be a positive integer (milliseconds).")
    if len(images) == 0:
        raise ValueError("images must be a non-empty sequence.")
    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
    if scan is not None:
        extent = scan.extent * 1e3 if getattr(scan, "extent") is not None else None
    else:
        extent = None
    im = ax.imshow(np.array(images[0]), animated=True, cmap=cmap, extent=extent)
    ax.set_xlabel("X (mm)")
    ax.set_ylabel("Z (mm)")

    def update(frame):
        im.set_array(np.array(images[frame]))
        return [im]

    ani = animation.FuncAnimation(
        fig,
        update,
        frames=len(images),
        blit=True,
        interval=interval,
    )
    plt.close(fig)
    fps = max(1, 1000 // interval)

    ani.save(path, writer="imagemagick", fps=fps)


def animate_volume_mip(
    volume,
    path,
    n_frames=60,
    interval=50,
    cmap="gray",
    axis=0,
    zoom=1.0,
):
    """Create an animated 3D MIP visualization rotating around the volume.

    Args:
        volume (tensor): 3D or 4D volume array.
            If 4D with shape (1, D, H, W), the first dim is squeezed.
        path (str): Output path for the gif file.
        n_frames (int): Number of frames for the 360-degree rotation.
        interval (int): Milliseconds between frames.
        cmap (str): Colormap name.
        axis (int): Axis of rotation: 0 (D/depth), 1 (H/vertical), or 2 (W/horizontal).
        zoom (float): Zoom factor. Values > 1 zoom in, < 1 zoom out.
    """
    # Squeeze batch dimension if present
    vol = np.squeeze(volume)
    if vol.ndim != 3:
        raise ValueError(f"Expected 3D volume after squeezing, got shape {vol.shape}")

    if axis not in (0, 1, 2):
        raise ValueError(f"axis must be 0, 1, or 2, got {axis}")

    # Normalize volume to 0-255 uint8
    vol = vol.astype(np.float32)
    vol = (vol - vol.min()) / (vol.max() - vol.min() + 1e-8)
    vol = (vol * 255).astype(np.uint8)

    # Transpose so rotation is always around axis 1 (vertical in output)
    # We rotate in the (axis0, axis2) plane, axis1 is the vertical output axis
    if axis == 0:
        # Rotate around original axis 0: (D, H, W) -> (H, D, W), rotate in (D, W) plane
        vol = np.transpose(vol, (1, 0, 2))
    elif axis == 1:
        # Rotate around original axis 1: (D, H, W) -> (D, H, W), rotate in (D, W) plane
        pass
    elif axis == 2:
        # Rotate around original axis 2: (D, H, W) -> (D, W, H), rotate in (D, H) plane
        vol = np.transpose(vol, (0, 2, 1))

    d, h, w = vol.shape

    # Generate rotation angles
    angles = np.linspace(0, 360, n_frames, endpoint=False)

    # Output size to fit rotated volume
    diag = int(np.ceil(np.sqrt(w**2 + d**2)))
    out_w = int(np.ceil(diag / zoom))
    out_h = int(np.ceil(h / zoom))

    cx, cz = w / 2, d / 2
    out_cx = out_w / 2

    # Pre-compute output coordinate grid for rotation plane
    x_out, z_out = np.meshgrid(
        (np.arange(out_w) - out_cx) / zoom,
        (np.arange(diag) - diag / 2) / zoom,
        indexing="xy",
    )

    # Get colormap LUT
    cmap_fn = plt.get_cmap(cmap)
    lut = (cmap_fn(np.arange(256))[:, :3] * 255).astype(np.uint8)

    def compute_rotated_mip(angle_deg):
        """Compute MIP by rotating in 2D plane then projecting."""
        angle_rad = np.deg2rad(angle_deg)
        cos_a, sin_a = np.cos(angle_rad), np.sin(angle_rad)

        # Inverse rotation to find source coordinates
        src_x = cos_a * x_out + sin_a * z_out + cx
        src_z = -sin_a * x_out + cos_a * z_out + cz

        # Validity mask
        valid = (src_x >= 0) & (src_x < w) & (src_z >= 0) & (src_z < d)

        # Nearest-neighbor indices
        xi = np.clip(np.rint(src_x).astype(np.int32), 0, w - 1)
        zi = np.clip(np.rint(src_z).astype(np.int32), 0, d - 1)

        # Gather all y slices at once: (diag, out_w, h)
        vals = vol[zi, :, xi]
        vals = np.where(valid[:, :, None], vals, 0)

        # MIP along depth (axis 0), result: (out_w, h)
        mip = np.max(vals, axis=0).T

        # Crop/pad to output height with zoom
        if out_h < h:
            start = (h - out_h) // 2
            mip = mip[start : start + out_h, :]
        elif out_h > h:
            pad_top = (out_h - h) // 2
            pad_bot = out_h - h - pad_top
            mip = np.pad(mip, ((pad_top, pad_bot), (0, 0)), mode="constant")

        return lut[mip]

    # Compute all frames
    frames = np.stack([compute_rotated_mip(angle) for angle in angles])

    # Save using zea's save_to_gif
    fps = max(1, 1000 // interval)
    save_to_gif(frames, path, fps=fps)



"""
Action selection strategies
===========================

These selection strategies implement a variety of policies for choosing which focused
transmit to fire next, potentially given some beliefs about the state of tissue,
as represented by ``particles``.

For a comprehensive example usage, see: :doc:`../notebooks/agent/agent_example`

All strategies are stateless, meaning that they do not maintain any internal state.
"""

from typing import Callable

import keras
from keras import ops

from zea.agent import masks
from zea.backend.autograd import AutoGrad
from zea.func import tensor
from zea.internal.registry import action_selection_registry


class MaskActionModel:
    """Base class for any action selection method that does masking."""

    def apply(self, action, observation):
        """Apply the action to the observation.

        Args:
            action (Tensor): The mask to be applied.
            observation (Tensor): The observation to which the action is applied.

        Returns:
            Tensor: The masked tensor
        """
        return observation * action


class LinesActionModel(MaskActionModel):
    """Base class for action selection methods that select lines."""

    def __init__(self, n_actions: int, n_possible_actions: int, img_width: int, img_height: int):
        """Initialize the LinesActionModel.

        Args:
            n_actions (int): The number of actions the agent can take.
            n_possible_actions (int): The number of possible actions.
            img_width (int): The width of the input image.
            img_height (int): The height of the input image.

        Attributes:
            stack_n_cols (int): The number of columns in the image that correspond
                to a single action.
        """
        super().__init__()
        self.n_actions = n_actions
        self.n_possible_actions = n_possible_actions
        self.img_width = img_width
        self.img_height = img_height

        stack_n_cols = self.img_width / self.n_possible_actions
        assert stack_n_cols.is_integer(), "Image width must be divisible by n_possible_actions."
        self.stack_n_cols = int(stack_n_cols)

    def lines_to_im_size(self, lines):
        """Convert k-hot-encoded line vectors to image size.

        Args:
            lines (Tensor): shape is (n_masks, n_possible_actions)

        Returns:
            Tensor: Masks of shape (n_masks, img_height, img_width)
        """
        return masks.lines_to_im_size(lines, (self.img_height, self.img_width))


@action_selection_registry(name="greedy_entropy")
class GreedyEntropy(LinesActionModel):
    """Greedy entropy action selection.

    Selects the max entropy line and reweights the entropy values around it,
    approximating the decrease in entropy that would occur from observing that line.

    The neighbouring values are decreased by a Gaussian function centered at the selected line.
    """

    def __init__(
        self,
        n_actions: int,
        n_possible_actions: int,
        img_width: int,
        img_height: int,
        mean: float = 0,
        std_dev: float = 1,
        num_lines_to_update: int = 5,
        entropy_sigma: float = 1.0,
        average_entropy_across_batch: bool = False,
    ):
        """Initialize the GreedyEntropy action selection model.

        Args:
            n_actions (int): The number of actions the agent can take.
            n_possible_actions (int): The number of possible actions.
            img_width (int): The width of the input image.
            img_height (int): The height of the input image.
            mean (float, optional): The mean of the RBF. Defaults to 0.
            std_dev (float, optional): The standard deviation of the RBF. Defaults to 1.
            num_lines_to_update (int, optional): The number of lines around the selected line
                to update. Must be odd.
            entropy_sigma (float, optional): The standard deviation of the Gaussian
                Mixture components used to approximate the posterior.
            average_entropy_across_batch (bool, optional): Whether to average entropy
                across the batch when selecting lines. This can be useful when
                selecting planes in 3D imaging, where the batch dimension represents
                a third spatial dimension. Defaults to False.
        """
        super().__init__(n_actions, n_possible_actions, img_width, img_height)

        # Number of samples must be odd so that the entropy
        # of the selected line is set to 0 once it's been selected.
        assert num_lines_to_update % 2 == 1, "num_samples must be odd."
        self.num_lines_to_update = num_lines_to_update
        self.average_entropy_across_batch = average_entropy_across_batch

        # see here what I mean by upside_down_gaussian:
        # https://colab.research.google.com/drive/1CQp_Z6nADzOFsybdiH5Cag0vtVZjjioU?usp=sharing
        upside_down_gaussian = lambda x: 1 - ops.exp(-0.5 * ((x - mean) / std_dev) ** 2)
        # Sample `num_lines_to_update` points symmetrically around the mean.
        # This can be tuned to determine how the entropy for neighbouring lines is updated
        # TODO: learn this function from training data
        points_to_evaluate = ops.linspace(
            mean - 2 * std_dev,
            mean + 2 * std_dev,
            self.num_lines_to_update,
        )
        self.upside_down_gaussian = upside_down_gaussian(points_to_evaluate)
        self.entropy_sigma = entropy_sigma
        self.reweight_entropies_around_line_vmap = tensor.vmap(self.reweight_entropies_around_line)

    @staticmethod
    def compute_pairwise_pixel_gaussian_error(particles, entropy_sigma=1.0):
        """Compute the pairwise pixelwise Gaussian error.

        This function computes the Gaussian error between each pair of pixels in the
        set of particles provided. This can be used to approximate the entropy of
        a Gaussian mixture model, where the particles are the means of the Gaussians.
        For more details see Section 4 here: https://arxiv.org/abs/2406.14388

        Args:
            particles (Tensor): Particles of shape (batch_size, n_particles, \*pixels)
            entropy_sigma (float, optional): The standard deviation of the Gaussian
                Mixture components used to approximate the posterior. Defaults to 1.0.

        Returns:
            Tensor: batch of pixelwise pairwise Gaussian errors,
            of shape (batch_size, n_particles, n_particles, \*pixels)
        """
        assert particles.shape[1] > 1, "The entropy cannot be approximated using a single particle."

        # TODO: I think we only need to compute the lower triangular
        # of this matrix, since it's symmetric
        squared_l2_error_matrices = (particles[:, :, None, ...] - particles[:, None, :, ...]) ** 2
        gaussian_error_per_pixel_i_j = ops.exp(
            -(squared_l2_error_matrices) / (2 * entropy_sigma**2)
        )
        # [batch_size, n_particles, n_particles, *pixels]
        return gaussian_error_per_pixel_i_j

    @staticmethod
    def compute_pixelwise_entropy(particles, entropy_sigma=1.0):
        """
        This function computes the entropy for each pixel using a Gaussian Mixture Model
        approximation of the posterior distribution.
        For more details see Section VI. B here: https://arxiv.org/pdf/2410.13310

        Args:
            particles (Tensor): Particles of shape (batch_size, n_particles, \*pixels)
            entropy_sigma (float, optional): The standard deviation of the Gaussian
                Mixture components used to approximate the posterior. Defaults to 1.0.

        Returns:
            Tensor: batch of entropies per pixel, of shape (batch_size, \*pixels)
        """
        n_particles = ops.shape(particles)[1]
        gaussian_error_per_pixel_stacked = GreedyEntropy.compute_pairwise_pixel_gaussian_error(
            particles, entropy_sigma
        )
        # sum out first dimension of (n_particles x n_particles) error matrix
        # [batch_size, n_particles, *pixels]
        pixelwise_entropy_sum_j = ops.sum(
            (1 / n_particles) * gaussian_error_per_pixel_stacked, axis=1
        )
        log_pixelwise_entropy_sum_j = ops.log(pixelwise_entropy_sum_j)
        # sum out second dimension of (n_particles x n_particles) error matrix
        # [batch_size, *pixels]
        pixelwise_entropy = -ops.sum((1 / n_particles) * log_pixelwise_entropy_sum_j, axis=1)
        return pixelwise_entropy

    def reweight_entropies_around_line(self, entropy_per_line, line_index):
        """Reweight the entropy around a selected line.

        This approximates the decrease in entropy that would occur from observing that line.
        It works by multiplying the entropy values around the selected line by an upside-down
        Gaussian function centered at the selected line, setting the entropy of the selected
        line to 0, and decreasing the entropies of neighbouring lines.

        .. note::

            This function is not compatible with the torch backend.
            See `Issue #268 <https://github.com/tue-bmd/zea/issues/268>`_

        Args:
            entropy_per_line (Tensor): Entropy per line of shape (n_possible_actions,)
            line_index (int): Index of the line with maximum entropy

        Returns:
            Tuple: The reweighted entropy per line, of shape (n_possible_actions,)
        """

        # Pad the entropy per line to allow for re-weighting with fixed
        # size RBF, which is necessary for tracing.
        padded_entropy_per_line = ops.pad(
            entropy_per_line,
            (self.num_lines_to_update // 2, self.num_lines_to_update // 2),
        )

        # Create the re-weighting vector
        # because the entropy per line has now been padded, the start index
        # of the set of lines to update is simply the line_index
        reweighting = ops.ones_like(padded_entropy_per_line)
        reweighting = ops.slice_update(
            reweighting,
            (line_index,),
            ops.cast(self.upside_down_gaussian, dtype=reweighting.dtype),
        )

        # Apply re-weighting to entropy values
        updated_entropy_per_line_padded = padded_entropy_per_line * reweighting
        updated_entropy_per_line = ops.slice(
            updated_entropy_per_line_padded,
            (self.num_lines_to_update // 2,),
            (self.n_possible_actions,),
        )
        return updated_entropy_per_line

    def sample(self, particles):
        """Sample the action using the greedy entropy method.

        Args:
            particles (Tensor): Particles of shape (batch_size, n_particles, height, width)

        Returns:
            Tuple[Tensor, Tensor]:
                - Newly selected lines as k-hot vectors, shaped (batch_size, n_possible_actions)
            - Masks of shape (batch_size, img_height, img_width)
        """

        pixelwise_entropy = self.compute_pixelwise_entropy(particles, self.entropy_sigma)

        # Sum over height to get column-wise entropy
        column_wise_entropy = ops.sum(pixelwise_entropy, axis=1)  # shape (b, w)

        # Reshape columns into lines and sum to get line-wise entropy
        linewise_entropy = ops.sum(
            ops.reshape(column_wise_entropy, (-1, self.n_possible_actions, self.stack_n_cols)),
            axis=2,
        )  # shape (b, n_possible_actions)

        # Optionally average entropy across batch dimension, which can be useful when selecting
        # planes in 3D imaging
        if self.average_entropy_across_batch:
            linewise_entropy = ops.expand_dims(ops.mean(linewise_entropy, axis=0), axis=0)

        # Greedily select best line, reweight entropies, and repeat
        all_selected_lines = []
        for _ in range(self.n_actions):
            max_entropy_line = ops.argmax(linewise_entropy, axis=1)
            linewise_entropy = self.reweight_entropies_around_line_vmap(
                linewise_entropy, max_entropy_line
            )
            all_selected_lines.append(max_entropy_line)

        selected_lines_k_hot = masks.indices_to_k_hot(
            ops.stack(all_selected_lines, axis=1), self.n_possible_actions
        )
        return selected_lines_k_hot, self.lines_to_im_size(selected_lines_k_hot)


@action_selection_registry(name="uniform_random")
class UniformRandomLines(LinesActionModel):
    """Uniform random lines action selection.

    Creates masks with uniformly randomly sampled lines.
    """

    def sample(self, batch_size=1, seed=None):
        """Sample the action using the uniform random method.

        Generates or updates an equispaced mask to sweep rightwards by one step across the image.

        Args:
            seed (int | SeedGenerator | jax.random.key, optional): Seed for random
                number generation. Defaults to None.

        Returns:
            Tuple[Tensor, Tensor]:
                - Newly selected lines as k-hot vectors, shaped (batch_size, n_possible_actions)
                - Masks of shape (batch_size, img_height, img_width)
        """
        selected_lines_batched = masks.random_uniform_lines(
            n_actions=self.n_actions,
            n_possible_actions=self.n_possible_actions,
            n_masks=batch_size,
            seed=seed,
        )
        return selected_lines_batched, self.lines_to_im_size(selected_lines_batched)


@action_selection_registry(name="equispaced")
class EquispacedLines(LinesActionModel):
    """Equispaced lines action selection.

    Creates masks with equispaced lines that sweep across
    the image.
    """

    def __init__(
        self,
        n_actions: int,
        n_possible_actions: int,
        img_width: int,
        img_height: int,
        assert_equal_spacing: bool = True,
    ):
        super().__init__(n_actions, n_possible_actions, img_width, img_height)

        self.assert_equal_spacing = assert_equal_spacing
        if self.assert_equal_spacing:
            masks._assert_equal_spacing(n_actions, n_possible_actions)

    def sample(self, current_lines=None, batch_size=1):
        """Sample the action using the equispaced method.

        Generates or updates an equispaced mask to sweep rightwards by one step across the image.

        Returns:
            Tensor: The mask of shape (batch_size, img_size, img_size)
        """
        if current_lines is None:
            return self.initial_sample_stateless(batch_size)
        else:
            return self.sample_stateless(current_lines)

    def initial_sample_stateless(self, batch_size=1):
        """Initial sample stateless.

        Generates a batch of initial equispaced line masks.

        Returns:
            Tuple[Tensor, Tensor]:
                - Selected lines as k-hot vectors, shaped (batch_size, n_possible_actions)
                - Masks of shape (batch_size, img_height, img_width)
        """
        initial_lines = masks.initial_equispaced_lines(
            self.n_actions,
            self.n_possible_actions,
            assert_equal_spacing=self.assert_equal_spacing,
        )
        initial_lines = ops.tile(initial_lines, (batch_size, 1))  # (batch_size, n_actions)
        return initial_lines, self.lines_to_im_size(initial_lines)

    def sample_stateless(self, current_lines):
        """Sample stateless.

        Updates an existing equispaced mask to sweep rightwards by one step across the image.

        Args:
            current_lines: Currently selected lines as k-hot vectors,
                shaped (batch_size, n_possible_actions)

        Returns:
            Tuple[Tensor, Tensor]:
                - Newly selected lines as k-hot vectors, shaped (batch_size, n_possible_actions)
                - Masks of shape (batch_size, img_height, img_width)
        """
        new_lines = masks.next_equispaced_lines(current_lines)
        return new_lines, self.lines_to_im_size(new_lines)


@action_selection_registry(name="covariance")
class CovarianceSamplingLines(LinesActionModel):
    """Covariance sampling action selection.

    This class models the line-to-line correlation to select the mask with the highest entropy.
    """

    def __init__(
        self,
        n_actions: int,
        n_possible_actions: int,
        img_width: int,
        img_height: int,
        seed: int = 42,
        n_masks: int = 200,
    ):
        """Initialize the CovarianceSamplingLines action selection model.

        Args:
            n_actions (int): The number of actions the agent can take.
            n_possible_actions (int): The number of possible actions.
            img_width (int): The width of the input image.
            img_height (int): The height of the input image.
            seed (int, optional): The seed for random number generation. Defaults to 42.
            n_masks (int, optional): The number of masks. Defaults to 200.

        Raises:
            AssertionError: If image width is not divisible by n_possible_actions.
        """
        super().__init__(n_actions, n_possible_actions, img_width, img_height)
        self.seed = keras.random.SeedGenerator(seed)
        self.n_masks = n_masks

    def random_uniform_lines(self, batch_size, seed=None):
        """Wrapper around `random_uniform_lines` function to use attributes from class."""
        lines = masks.random_uniform_lines(
            self.n_actions,
            self.n_possible_actions,
            batch_size * self.n_masks,
            seed=self.seed if seed is None else seed,
        )
        return ops.reshape(lines, [self.n_masks, batch_size, self.n_possible_actions])

    def sample(self, particles, seed=None):
        """Sample the action using the covariance sampling method.

        Args:
            particles (Tensor): Particles of shape (batch_size, n_particles, h, w)
            seed (int | SeedGenerator | jax.random.key, optional): Seed for random number
                generation. Defaults to None.

        Returns:
            Tuple[Tensor, Tensor]:
                - Newly selected lines as k-hot vectors, shaped (batch_size, n_possible_actions)
                - Masks of shape (batch_size, img_height, img_width)
        """
        batch_size, n_particles, rows, _ = ops.shape(particles)

        # [batch_size, rows, cols, n_particles]
        particles = ops.transpose(particles, (0, 2, 3, 1))

        # [batch_size, rows * stack_n_cols, n_possible_actions, n_particles]
        shape = [
            batch_size,
            rows * self.stack_n_cols,
            self.n_possible_actions,
            n_particles,
        ]
        particles = ops.reshape(particles, shape)

        # [batch_size, rows * stack_n_cols, n_possible_actions, n_possible_actions]
        cov_matrix = tensor.batch_cov(particles)

        # Sum over the row dimension [batch_size, n_possible_actions, n_possible_actions]
        cov_matrix = ops.sum(cov_matrix, axis=1)

        # Generate random lines [n_masks, batch_size, n_possible_actions]
        lines = self.random_uniform_lines(batch_size, seed=seed)

        # Make matrix masks [n_masks, batch_size, n_possible_actions, n_possible_actions]
        reshaped_lines = ops.repeat(lines[..., None], self.n_possible_actions, axis=-1)
        bool_masks = ops.logical_and(reshaped_lines, ops.swapaxes(reshaped_lines, -1, -2))

        # Subsample the covariance matrix with random lines
        def subsample_with_mask(mask):
            """Subsample the covariance matrix with a single mask."""
            subsampled_cov_matrix = tensor.boolean_mask(
                cov_matrix, mask, size=batch_size * self.n_actions**2
            )
            return ops.reshape(subsampled_cov_matrix, [batch_size, self.n_actions, self.n_actions])

        # [n_masks, batch_size, cols, cols]
        subsampled_cov_matrices = ops.vectorized_map(subsample_with_mask, bool_masks)

        # [n_masks, batch_size, 1]
        entropies = ops.logdet(subsampled_cov_matrices)[..., None]

        # [1, batch_size, 1]
        best_mask_index = ops.argmax(entropies, axis=0, keepdims=True)

        # [batch_size, n_possible_actions]
        best_mask = ops.take_along_axis(lines, best_mask_index, axis=0)
        best_mask = ops.squeeze(best_mask, axis=0)

        return best_mask, self.lines_to_im_size(best_mask)


class TaskBasedLines(GreedyEntropy):
    """Task-based line selection for maximizing information gain.

    This action selection strategy chooses lines to maximize information gain with respect
    to a downstream task outcome. It uses gradient-based saliency to identify which image
    regions contribute most to task uncertainty, then selects lines accordingly.
    """

    def __init__(
        self,
        n_actions: int,
        n_possible_actions: int,
        img_width: int,
        img_height: int,
        downstream_task_function: Callable,
        mean: float = 0,
        std_dev: float = 1,
        num_lines_to_update: int = 5,
        **kwargs,
    ):
        """Initialize the TaskBasedLines action selection model.

        Args:
            n_actions (int): The number of actions the agent can take.
            n_possible_actions (int): The number of possible actions (line positions).
            img_width (int): The width of the input image.
            img_height (int): The height of the input image.
            downstream_task_function (Callable): A differentiable function that takes a
                batch of inputs and produces scalar outputs. This represents the downstream
                task for which information gain should be maximized.
            mean (float, optional): The mean of the RBF used for reweighting. Defaults to 0.
            std_dev (float, optional): The standard deviation of the RBF used for reweighting.
                Defaults to 1.
            num_lines_to_update (int, optional): The number of lines around the selected line
                to update during reweighting. Must be odd. Defaults to 5.
            **kwargs: Additional keyword arguments passed to the parent class.
        """
        super().__init__(
            n_actions,
            n_possible_actions,
            img_width,
            img_height,
            mean,
            std_dev,
            num_lines_to_update,
        )
        self.downstream_task_function = downstream_task_function

    def compute_output_and_saliency_propagation(self, particles):
        """Compute saliency-weighted posterior variance for task-based selection.

        This method computes how much each pixel contributes to the variance of the
        downstream task output. It uses automatic differentiation to compute gradients
        of the task function with respect to each particle, then weights the posterior
        variance by the squared mean gradient.

        Args:
            particles (Tensor): Particles of shape (batch_size, n_particles, height, width)
                representing the posterior distribution over images.

        Returns:
            Tensor: Pixelwise contribution to downstream task variance,
                of shape (batch_size, height, width). Higher values indicate pixels
                that contribute more to task uncertainty.
        """
        autograd = AutoGrad()

        autograd.set_function(self.downstream_task_function)
        downstream_grad_and_value_fn = autograd.get_gradient_and_value_jit_fn()
        jacobian, _ = ops.vectorized_map(
            lambda p: ops.vectorized_map(
                downstream_grad_and_value_fn,
                p,
            ),
            particles,
        )

        posterior_variance = ops.var(particles, axis=1)
        mean_jacobian = ops.mean(jacobian, axis=1)
        return posterior_variance * (mean_jacobian**2)

    def sum_neighbouring_columns_into_n_possible_actions(self, full_linewise_salience):
        """Aggregate column-wise saliency into line-wise saliency scores.

        This method groups neighboring columns together to create saliency scores
        for each possible line action. Since each line action may correspond to
        multiple image columns, this aggregation is necessary to match the action space.

        Args:
            full_linewise_salience (Tensor): Saliency values for each column,
                of shape (batch_size, full_image_width).

        Returns:
            Tensor: Aggregated saliency scores for each possible action,
                of shape (batch_size, n_possible_actions).

        Raises:
            AssertionError: If the image width is not evenly divisible by n_possible_actions.
        """
        batch_size = ops.shape(full_linewise_salience)[0]
        full_image_width = ops.shape(full_linewise_salience)[1]
        assert full_image_width % self.n_possible_actions == 0, (
            "n_possible_actions must divide evenly into image width"
        )
        cols_per_action = full_image_width // self.n_possible_actions
        stacked_linewise_salience = ops.reshape(
            full_linewise_salience,
            (batch_size, self.n_possible_actions, cols_per_action),
        )
        return ops.sum(stacked_linewise_salience, axis=2)

    def sample(self, particles):
        """Sample actions using task-based information gain maximization.

        This method computes which lines would provide the most information about
        the downstream task by:
        1. Computing pixelwise contribution to task variance using gradients
        2. Aggregating contributions into line-wise scores
        3. Greedily selecting lines with highest contribution scores
        4. Reweighting scores around selected lines (inherited from GreedyEntropy)

        Args:
            particles (Tensor): Particles representing the posterior distribution,
                of shape (batch_size, n_particles, height, width).

        Returns:
            Tuple[Tensor, Tensor, Tensor]:
                - selected_lines_k_hot: Selected lines as k-hot vectors,
                  shaped (batch_size, n_possible_actions)
                - masks: Binary masks of shape (batch_size, img_height, img_width)
                - pixelwise_contribution_to_var_dst: Pixelwise contribution to downstream
                  task variance, of shape (batch_size, height, width)

        Note:
            Unlike the parent GreedyEntropy class, this method returns an additional
            tensor containing the pixelwise contribution scores for analysis.
        """
        pixelwise_contribution_to_var_dst = self.compute_output_and_saliency_propagation(particles)
        linewise_contribution_to_var_dst = ops.sum(pixelwise_contribution_to_var_dst, axis=1)
        actionwise_contribution_to_var_dst = self.sum_neighbouring_columns_into_n_possible_actions(
            linewise_contribution_to_var_dst
        )

        # Greedily select best line, reweight entropies, and repeat
        all_selected_lines = []
        for _ in range(self.n_actions):
            max_contribution_line = ops.argmax(actionwise_contribution_to_var_dst, axis=1)
            actionwise_contribution_to_var_dst = self.reweight_entropies_around_line_vmap(
                actionwise_contribution_to_var_dst, max_contribution_line
            )
            all_selected_lines.append(max_contribution_line)

        selected_lines_k_hot = masks.indices_to_k_hot(
            ops.stack(all_selected_lines, axis=1), self.n_possible_actions
        )
        return (
            selected_lines_k_hot,
            self.lines_to_im_size(selected_lines_k_hot),
            pixelwise_contribution_to_var_dst,
        )



"""Data subpackage for working with the ``zea`` data format.

This subpackage provides core classes and utilities for working with the zea data format,
including file and dataset access, validation, and data loading. For more information on the
``zea`` data format, see :doc:`../data-acquisition`.

Main classes
------------

- :class:`zea.data.File` -- Open and access a single zea HDF5 data file.
- :class:`zea.data.Dataset` -- Manage and iterate over a collection of zea data files.

See the data notebook for a more detailed example: :doc:`../notebooks/data/zea_data_example`

Examples usage
^^^^^^^^^^^^^^

.. doctest::

    >>> from zea import File, Dataset

    >>> # Work with a single file
    >>> path_to_file = (
    ...     "hf://zeahub/picmus/database/experiments/contrast_speckle/"
    ...     "contrast_speckle_expe_dataset_iq/contrast_speckle_expe_dataset_iq.hdf5"
    ... )

    >>> with File(path_to_file, mode="r") as file:
    ...     # file.summary()
    ...     data = file.load_data("raw_data", indices=[0])
    ...     scan = file.scan()
    ...     probe = file.probe()

    >>> # Work with a dataset (folder or list of files)
    >>> dataset = Dataset("hf://zeahub/picmus")
    >>> files = []
    >>> for file in dataset:
    ...     files.append(file)  # process each file as needed
    >>> dataset.close()

"""  # noqa: E501

from .convert.camus import sitk_load
from .data_format import (
    DatasetElement,
    generate_example_dataset,
    generate_zea_dataset,
    validate_input_data,
)
from .dataloader import H5Generator
from .datasets import Dataset, Folder
from .file import File, load_file



"""Agent subpackage for closing action-perception loop in ultrasound imaging.

The ``zea.agent`` subpackage provides tools and utilities for implementing **cognitive ultrasound**
imaging via active perception. This module enables intelligent, adaptive transmit design, where
acquisition decisions are informed by the current belief state about the imaged tissue.

Overview
========

Active perception in ultrasound involves iteratively:

1. **Perceiving** the current state of tissue from acquired measurements
2. **Selecting transmit actions** based on the current beliefs about the tissue state
3. **Acquiring** new data and looping back to the perception step

The :mod:`zea.agent` module provides the building blocks for implementing such perception-action
loops.

.. note::
    The functions currently available implement selection strategies for *focused transmit actions*. Development of action selection functions for more general transmit schemes is currently a work-in-progress.

Action Selection Strategies
===========================

Action selection strategies determine which transmits to fire next, given some belief
about the tissue state.

See the following dropdown for a list of available action selection strategies:

.. dropdown:: **Available strategies**

    - :class:`zea.agent.selection.GreedyEntropy`: Selects lines that maximize entropy reduction.
    - :class:`zea.agent.selection.UniformRandomLines`: Randomly samples scan lines with uniform probability.
    - :class:`zea.agent.selection.EquispacedLines`: Selects equispaced lines that sweep across the image.
    - :class:`zea.agent.selection.CovarianceSamplingLines`: Models line-to-line correlation to select masks with highest entropy.
    - :class:`zea.agent.selection.TaskBasedLines`: Selects lines to maximize information gain with respect to a downstream task.

Basic Usage
===========

.. doctest::

    >>> import zea
    >>> import numpy as np

    >>> agent = zea.agent.selection.GreedyEntropy(
    ...     n_actions=7,
    ...     n_possible_actions=112,
    ...     img_width=112,
    ...     img_height=112,
    ... )

    >>> # (batch, samples, height, width)
    >>> particles = np.random.rand(1, 10, 112, 112)
    >>> lines, mask = agent.sample(particles)

Masks
=====

The :mod:`zea.agent.masks` module provides utilities for converting action representations
(e.g., selected line indices) to image-sized masks that can be applied to observations.

Example Notebooks
=================

We provide example notebooks demonstrating perception-action loops in practice, as companions to recently published papers on the topic:

**Patient-Adaptive Echocardiography**

This tutorial implements a basic perception-action loop using diffusion models for
perception-as-inference and greedy entropy minimization for action selection.

- :doc:`../notebooks/agent/agent_example`

  - Uses :class:`~zea.agent.selection.GreedyEntropy` to select informative scan lines
  - Demonstrates iterative belief refinement with sparse acquisitions
  - Visualizes the reconstruction process over multiple acquisition steps

**Task-Based Transmit Beamforming**

This tutorial implements a task-driven perception-action loop where acquisition decisions
are optimized to gain information about a specific downstream task.

- :doc:`../notebooks/agent/task_based_perception_action_loop`

  - Uses :class:`~zea.agent.selection.TaskBasedLines` for task-aware line selection
  - Computes saliency maps via uncertainty propagation through downstream task models
  - Demonstrates how to integrate domain-specific measurement tasks
"""

from . import masks, selection



"""EchoNetLVH model for segmentation of PLAX view cardiac ultrasound.

To try this model, simply load one of the available presets:

.. doctest::

    >>> from zea.models.echonetlvh import EchoNetLVH

    >>> model = EchoNetLVH.from_preset("echonetlvh")

.. important::
    This is a ``zea`` implementation of the model.
    For the original paper and code, see `here <https://echonet.github.io/lvh/>`_.

    Duffy, Grant, et al.
    "High-throughput precision phenotyping of left ventricular hypertrophy with cardiovascular deep learning."
    *JAMA cardiology 7.4 (2022): 386-395*

.. seealso::
    A tutorial notebook where this model is used:
    :doc:`../notebooks/agent/task_based_perception_action_loop`.

"""  # noqa: E501

import numpy as np
from keras import ops

from zea.func.tensor import translate
from zea.internal.registry import model_registry
from zea.models.base import BaseModel
from zea.models.deeplabv3 import DeeplabV3Plus
from zea.models.preset_utils import register_presets
from zea.models.presets import echonet_lvh_presets


@model_registry(name="echonetlvh")
class EchoNetLVH(BaseModel):
    """
    EchoNet Left Ventricular Hypertrophy (LVH) model for echocardiogram analysis.

    This model performs semantic segmentation on echocardiogram images to identify
    key anatomical landmarks for measuring left ventricular wall thickness:

    - **LVPWd_1**: Left Ventricular Posterior Wall point 1
    - **LVPWd_2**: Left Ventricular Posterior Wall point 2
    - **IVSd_1**: Interventricular Septum point 1
    - **IVSd_2**: Interventricular Septum point 2

    The model outputs 4-channel logits corresponding to heatmaps for each landmark.

    For more information, see the original project page:
    https://echonet.github.io/lvh/
    """

    def __init__(self, **kwargs):
        """
        Initialize the EchoNetLVH model.

        Args:
            **kwargs: Additional keyword arguments passed to BaseModel
        """
        super().__init__(**kwargs)

        # Pre-computed coordinate grid for efficient processing
        self.coordinate_grid = ops.stack(
            ops.cast(ops.convert_to_tensor(np.indices((224, 224))), "float32"), axis=-1
        )

        # Initialize the underlying segmentation network
        self.network = DeeplabV3Plus(image_shape=(224, 224, 3), num_classes=4)

    def call(self, inputs):
        """
        Forward pass of the model.

        Args:
            inputs (Tensor): Input images of shape [B, H, W, C]. They should
                be scan converted, with pixel values in range [0, 255].

        Returns:
            Tensor: Logits of shape [B, H, W, 4] with 4 channels for each landmark
        """
        assert len(ops.shape(inputs)) == 4

        # Store original dimensions for output resizing
        original_size = ops.shape(inputs)[1:3]

        # Resize to network input size
        inputs_resized = ops.image.resize(inputs, size=(224, 224))

        # Get network predictions
        logits = self.network(inputs_resized)

        # Resize logits back to original input dimensions
        logits_output = ops.image.resize(logits, original_size)
        return logits_output

    def extract_key_points_as_indices(self, logits):
        """
        Extract key point coordinates from logits using center-of-mass calculation.

        Args:
            logits (Tensor): Model output logits of shape [B, H, W, 4]

        Returns:
            Tensor: Key point coordinates of shape [B, 4, 2] where each point
                   is in (x, y) format
        """
        # Create coordinate grid for the current logit dimensions
        input_shape = ops.shape(logits)[1:3]
        input_space_coordinate_grid = ops.stack(
            ops.cast(ops.convert_to_tensor(np.indices(input_shape)), "float32"), axis=-1
        )

        # Transpose logits to [B, 4, H, W] for vectorized processing
        logits_batchified = ops.transpose(logits, (0, 3, 1, 2))

        # Extract expected coordinates for each channel
        return ops.flip(
            ops.vectorized_map(
                lambda logit: self.expected_coordinate(logit, input_space_coordinate_grid),
                logits_batchified,
            ),
            axis=-1,  # Flip to convert from (y, x) to (x, y)
        )

    def expected_coordinate(self, mask, coordinate_grid=None):
        """
        Compute the expected coordinate (center-of-mass) of a heatmap.

        This implements a differentiable version of taking the max of a heatmap
        by computing the weighted average of coordinates.

        Reference: https://arxiv.org/pdf/1711.08229

        Args:
            mask (Tensor): Heatmap of shape [B, H, W]
            coordinate_grid (Tensor, optional): Grid of coordinates. If None,
                                              uses self.coordinate_grid

        Returns:
            Tensor: Expected coordinates of shape [B, 2] in (x, y) format
        """
        if coordinate_grid is None:
            coordinate_grid = self.coordinate_grid

        # Ensure mask values are non-negative and normalized
        mask_clipped = ops.clip(mask, 0, None)
        mask_normed = mask_clipped / ops.max(mask_clipped)

        def safe_normalize(m):
            mask_sum = ops.sum(m)
            return ops.where(mask_sum > 0, m / mask_sum, m)

        coordinate_probabilities = ops.map(safe_normalize, mask_normed)

        # Add dimension for broadcasting with coordinate grid
        coordinate_probabilities = ops.expand_dims(coordinate_probabilities, axis=-1)

        # Compute weighted average of coordinates
        expected_coordinate = ops.sum(
            ops.expand_dims(coordinate_grid, axis=0) * coordinate_probabilities,
            axis=(1, 2),
        )

        # Flip to convert from (y, x) to (x, y) format for euclidean distance calculation
        return ops.flip(expected_coordinate, axis=-1)

    def overlay_labels_on_image(self, image, label, alpha=0.5):
        """
        Overlay predicted heatmaps and connecting lines on the input image.

        Args:
            image (Tensor): Input image of shape [H, W] or [H, W, C]
            label (Tensor): Predicted logits of shape [H, W, 4]
            alpha (float): Blending factor for overlay (0=transparent, 1=opaque)

        Returns:
            ndarray: Image with overlaid heatmaps and measurements of shape [H, W, 3]
        """
        try:
            import cv2

        except ImportError as exc:
            raise ImportError(
                "OpenCV is required for `EchoNetLVH.overlay_labels_on_image`. "
                "Please install it with 'pip install opencv-python' or "
                "'pip install opencv-python-headless'."
            ) from exc

        # Color scheme for each landmark
        overlay_colors = np.array(
            [
                [1, 1, 0],  # Yellow (LVPWd_X1)
                [1, 0, 1],  # Magenta (LVPWd_X2)
                [0, 1, 1],  # Cyan (IVSd_X1)
                [0, 1, 0],  # Green (IVSd_X2)
            ],
        )

        # Convert to numpy and ensure RGB format
        image = ops.convert_to_numpy(image)
        label = ops.convert_to_numpy(label)

        if image.ndim == 2:
            image = np.stack([image] * 3, axis=-1)
        elif image.shape[-1] == 1:
            image = np.repeat(image, 3, axis=-1)
        else:
            image = image.copy()

        # Normalize each channel to [0, 1] for proper visualization
        label = np.clip(label, 0, None)
        for ch in range(label.shape[-1]):
            max_val = np.max(label[..., ch])
            if max_val > 0:
                label[..., ch] = label[..., ch] / max_val

        # Initialize overlay and tracking variables
        overlay = np.zeros_like(image, dtype=np.float32)
        centers = []

        # Process each landmark channel
        for ch in range(4):
            # Square the mask to enhance peak responses
            mask = label[..., ch] ** 2
            color = overlay_colors[ch]

            # Find center of mass for this channel
            center_coords = self.expected_coordinate(ops.expand_dims(mask, axis=0))
            center_x = ops.convert_to_numpy(center_coords[0, 0])
            center_y = ops.convert_to_numpy(center_coords[0, 1])

            # Bounds check before conversion to int
            if 0 <= center_x < image.shape[1] and 0 <= center_y < image.shape[0]:
                center = (int(center_x), int(center_y))
            else:
                center = None

            if center is not None:
                # Blend heatmap with overlay
                mask_alpha = mask * alpha
                for c in range(3):
                    overlay[..., c] += mask_alpha * color[c]
            centers.append(center)

        # Draw connecting lines between consecutive landmarks
        for i in range(3):
            pt1, pt2 = centers[i], centers[i + 1]
            if pt1 is not None and pt2 is not None:
                color = tuple(int(x) for x in overlay_colors[i])

                # Create line mask
                line_mask = np.zeros(image.shape[:2], dtype=np.uint8)
                cv2.line(line_mask, pt1, pt2, color=1, thickness=2)

                # Apply line to overlay
                for c in range(3):
                    overlay[..., c][line_mask.astype(bool)] = color[c] * alpha

        # Blend overlay with original image
        overlay = np.clip(overlay, 0, 1)
        out = image.astype(np.float32)
        blend_mask = np.any(overlay > 0.02, axis=-1)
        out[blend_mask] = (1 - alpha) * out[blend_mask] + overlay[blend_mask]

        return np.clip(out, 0, 1)

    def visualize_logits(self, images, logits):
        """
        Create visualization of model predictions overlaid on input images.

        Args:
            images (Tensor): Input images of shape [B, H, W, C]
            logits (Tensor): Model predictions of shape [B, H, W, 4]

        Returns:
            Tensor: Images with overlaid predictions of shape [B, H, W, 3]
        """
        # Store original dimensions for final output
        original_size = ops.shape(images)[1:3]

        # Resize to standard processing size
        images_resized = ops.image.resize(images, size=(224, 224), interpolation="nearest")
        logits_resized = ops.image.resize(logits, size=(224, 224), interpolation="nearest")

        # Normalize images to [0, 1] range
        images_clipped = ops.clip(images_resized, 0, 255)
        images = translate(images_clipped, range_from=(0, 255), range_to=(0, 1))

        # Generate overlays for each image in the batch
        images_with_overlay = []
        for img, logit_heatmap in zip(images, logits_resized):
            overlay = self.overlay_labels_on_image(img, logit_heatmap)
            images_with_overlay.append(overlay)

        # Stack results and resize back to original dimensions
        images_with_overlay = np.stack(images_with_overlay, axis=0)
        return ops.image.resize(images_with_overlay, original_size)


# Register model presets
register_presets(echonet_lvh_presets, EchoNetLVH)
